# Copyright 2018--2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as onp
import pytest
import torch as pt

import sockeye.constants as C
import sockeye.transformer_pt
from sockeye.layers_pt import prepare_source_length_mask


def test_auto_regressive_bias_dtype():
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.transformer

    block = sockeye.transformer.AutoRegressiveBias()
    block.initialize()
    length = 10
    dtype = 'float32'
    data = np.ones((2, length, 10), dtype=dtype)
    bias = block(data)
    assert bias.dtype == np.float32

    dtype = 'float16'
    block.cast(dtype)
    bias = block(data.astype(dtype))
    assert bias.dtype == np.float16
    assert bias.min().item() == -C.LARGE_VALUES[dtype]


def test_auto_regressive_bias_output():
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.transformer

    block = sockeye.transformer.AutoRegressiveBias()
    block.initialize()
    length = 2
    data = np.ones((2, length, 10), dtype='float32')
    bias = block(data)

    expected = np.array([[0.0, -1.0e8], [0.0, 0.0]]).reshape((1, 2, 2))
    onp.testing.assert_array_equal(bias, expected)


@pytest.mark.parametrize('use_glu', [(False), (True)])
def test_transformer_feed_forward(use_glu):
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.transformer
    block = sockeye.transformer.TransformerFeedForward(num_hidden=2,
                                                       num_model=2,
                                                       act_type=C.RELU,
                                                       dropout=0.1,
                                                       dtype=C.DTYPE_FP32,
                                                       use_glu=use_glu)
    block.initialize()
    block.hybridize()

    data = np.ones((1, 10, 2), dtype=C.DTYPE_FP32)
    block(data)


@pytest.mark.parametrize('use_glu', [(False), (True)])
def test_pt_transformer_feed_forward(use_glu):
    block = sockeye.transformer_pt.PyTorchTransformerFeedForward(num_hidden=2,
                                                                 num_model=2,
                                                                 act_type=C.RELU,
                                                                 dropout=0.1,
                                                                 use_glu=use_glu)

    data = pt.ones(1, 10, 2)
    block(data)


@pytest.mark.parametrize('use_glu', [(False), (True)])
def test_mx_pt_eq_transformer_feed_forward(use_glu):
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.transformer

    b_mx = sockeye.transformer.TransformerFeedForward(num_hidden=4,
                                                      num_model=2,
                                                      act_type=C.RELU,
                                                      dropout=0.0,
                                                      dtype=C.DTYPE_FP32,
                                                      use_glu=use_glu)
    b_mx.initialize()

    b_pt = sockeye.transformer_pt.PyTorchTransformerFeedForward(num_hidden=4,
                                                                num_model=2,
                                                                act_type=C.RELU,
                                                                dropout=0.0,
                                                                use_glu=use_glu)
    b_pt.weights_from_mxnet_block(b_mx)

    result_mx = b_mx(np.ones((2, 2, 2))).asnumpy()
    result_pt = b_pt(pt.ones(2, 2, 2)).detach().numpy()

    assert np.allclose(result_mx, result_pt)


@pytest.mark.parametrize('length', [1, 10, 100])
def test_pt_autoregressive_mask(length):
    x_pt = pt.zeros(2, length, 32)
    b_pt = sockeye.transformer_pt.AutoRegressiveMask()
    result_pt = b_pt(x_pt).detach()

    assert result_pt.dtype == pt.bool
    assert result_pt.size() == (1, length, length)


@pytest.mark.parametrize('sequence', ['rn', 'nr', 'r', 'n', ''])  # not testing dropout
def test_mx_pt_eq_transformer_process_block(sequence):
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.transformer

    num_hidden = 32
    x_mx = np.random.uniform(0, 1, (2, 10, num_hidden))
    prev_mx = np.random.uniform(0, 1, (2, 10, num_hidden))
    x_pt = pt.as_tensor(x_mx.asnumpy())
    prev_pt = pt.as_tensor(prev_mx.asnumpy())

    b_mx = sockeye.transformer.TransformerProcessBlock(sequence, 0.0, num_hidden)
    b_mx.initialize()
    b_pt = sockeye.transformer_pt.PyTorchTransformerProcessBlock(sequence, 0.0, num_hidden)
    b_pt.weights_from_mxnet_block(b_mx)

    result_mx = b_mx(x_mx, prev_mx).asnumpy()
    result_pt = b_pt(x_pt, prev_pt).detach().numpy()

    assert np.allclose(result_mx, result_pt, atol=1e-06)


@pytest.mark.parametrize('batch_size, input_len, model_size, heads, ff_hidden, num_layers, use_lhuc, '
                         'preprocess_sequence, postprocess_sequence, use_glu',
                         [
                             (1, 100, 512, 8, 1024, 1, False, 'n', 'r', False),
                             (5, 25, 128, 4, 32, 1, True, 'n', 'nr', True),
                         ])
def test_mx_pt_eq_transformer_encoder_block(batch_size, input_len, model_size, heads,
                                            ff_hidden, num_layers, use_lhuc, preprocess_sequence,
                                            postprocess_sequence, use_glu):
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.transformer

    config = sockeye.transformer.TransformerConfig(
        model_size=model_size,
        attention_heads=heads,
        feed_forward_num_hidden=ff_hidden,
        act_type=C.RELU,
        num_layers=num_layers,
        dropout_attention=0.0,
        dropout_act=0.0,
        dropout_prepost=0.0,
        positional_embedding_type=C.FIXED_POSITIONAL_EMBEDDING,
        preprocess_sequence=preprocess_sequence,
        postprocess_sequence=postprocess_sequence,
        max_seq_len_source=100,
        max_seq_len_target=100,
        decoder_type=C.TRANSFORMER_TYPE,
        use_lhuc=use_lhuc,
        depth_key_value=model_size,
        use_glu=use_glu)

    data_mx = np.random.uniform(0, 1, (batch_size, input_len, model_size))
    data_pt = pt.as_tensor(data_mx.asnumpy())
    valid_lengths_mx = np.random.randint(0, input_len, (batch_size,))
    valid_lengths_pt = pt.as_tensor(valid_lengths_mx.asnumpy())

    att_valid_lengths_mx = sockeye.layers.prepare_source_valid_lengths(valid_lengths_mx, data_mx, num_heads=heads)
    source_length_mask_pt = sockeye.layers_pt.prepare_source_length_mask(valid_lengths_pt,
                                                                         heads=heads, max_length=data_pt.size()[1])

    # time-major as done in the Transformer encoder
    data_mx = np.transpose(data_mx, axes=(1, 0, 2))
    data_pt = data_pt.permute(1, 0, 2)

    b_mx = sockeye.transformer.TransformerEncoderBlock(config, dtype=C.DTYPE_FP32)
    b_mx.initialize()
    r_mx = b_mx(data_mx, att_valid_lengths_mx).asnumpy()

    b_pt = sockeye.transformer_pt.PyTorchTransformerEncoderBlock(config)
    b_pt.weights_from_mxnet_block(b_mx)
    r_pt = b_pt(data_pt, source_length_mask_pt).detach().numpy()

    assert np.allclose(r_mx, r_pt, atol=1e-05)


@pytest.mark.parametrize('batch_size, source_input_len, target_input_len, model_size, heads, ff_hidden, '
                         'num_layers, use_lhuc, preprocess_sequence, postprocess_sequence, use_glu, '
                         'inference_only, decoder_type',
                         [
                             (1, 100, 100, 512, 8, 1024, 1, False, 'n', 'r', False, False, C.TRANSFORMER_TYPE),
                             (5, 23, 25, 128, 4, 32, 1, True, 'n', 'nr', True, False, C.TRANSFORMER_TYPE),
                             (1, 55, 54, 32, 2, 64, 1, False, '', '', False, False, C.SSRU_TRANSFORMER),
                             (1, 55, 54, 32, 2, 64, 1, False, '', '', False, True, C.SSRU_TRANSFORMER)
                         ])
def test_mx_pt_eq_transformer_decoder_block(batch_size, source_input_len, target_input_len, model_size, heads,
                                            ff_hidden, num_layers, use_lhuc, preprocess_sequence,
                                            postprocess_sequence, use_glu, inference_only, decoder_type):
    mxnet = pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.transformer
    import sockeye.layers

    pt.manual_seed(13)
    mxnet.random.seed(13)
    config = sockeye.transformer.TransformerConfig(
        model_size=model_size,
        attention_heads=heads,
        feed_forward_num_hidden=ff_hidden,
        act_type=C.RELU,
        num_layers=num_layers,
        dropout_attention=0.0,
        dropout_act=0.0,
        dropout_prepost=0.0,
        positional_embedding_type=C.FIXED_POSITIONAL_EMBEDDING,
        preprocess_sequence=preprocess_sequence,
        postprocess_sequence=postprocess_sequence,
        max_seq_len_source=100,
        max_seq_len_target=100,
        decoder_type=decoder_type,
        use_lhuc=use_lhuc,
        depth_key_value=model_size,
        use_glu=use_glu)

    target_mx = np.random.uniform(0, 1, (batch_size, target_input_len, model_size))
    target_pt = pt.tensor(target_mx.asnumpy())

    autoregr_bias_mx = sockeye.transformer.AutoRegressiveBias()
    autoregr_bias_mx.initialize()
    autoregr_bias_pt = sockeye.transformer_pt.AutoRegressiveMask()
    target_bias_mx = autoregr_bias_mx(target_mx)
    target_bias_pt = autoregr_bias_pt(target_pt).detach()

    source_mx = np.random.uniform(0, 1, (batch_size, source_input_len, model_size))
    source_mx = np.transpose(source_mx, axes=(1, 0, 2))
    source_pt = pt.tensor(source_mx.asnumpy())

    source_lengths_mx = np.random.randint(0, source_input_len, (batch_size,))
    source_lengths_pt = pt.tensor(source_lengths_mx.asnumpy())
    source_lengths_mx = sockeye.layers.prepare_source_valid_lengths(source_lengths_mx, target_mx, heads)
    source_max_len = source_pt.size()[0]
    source_length_mask_pt = prepare_source_length_mask(source_lengths_pt, heads, source_max_len)

    target_mx = np.transpose(target_mx, axes=(1, 0, 2))
    target_pt = target_pt.permute(1, 0, 2)

    b_mx = sockeye.transformer.TransformerDecoderBlock(config, inference_only, C.DTYPE_FP32)
    b_mx.initialize()
    autoregr_states_mx = np.zeros(b_mx.get_states_shape(batch_size))
    new_target_mx, new_states_mx = b_mx(target_mx, target_bias_mx, source_mx, source_lengths_mx, autoregr_states_mx)
    new_target_mx = new_target_mx.asnumpy()
    new_states_mx = [s.asnumpy() for s in new_states_mx]

    b_pt = sockeye.transformer_pt.PyTorchTransformerDecoderBlock(config, inference_only)
    autoregr_states_pt = pt.zeros(*b_pt.get_states_shape(batch_size))
    b_pt.weights_from_mxnet_block(b_mx)
    new_target_pt, new_states_pt = b_pt(target_pt, target_bias_pt, source_pt, source_length_mask_pt, autoregr_states_pt)
    new_target_pt = new_target_pt.detach().numpy()
    new_states_pt = [s.detach().numpy() for s in new_states_pt]

    assert len(new_states_mx) == len(new_states_pt)
    assert np.allclose(new_target_mx, new_target_pt, atol=1e-05)
    assert np.allclose(new_states_mx, new_states_pt, atol=1e-05)



# Copyright 2017--2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as onp
import pytest
import torch as pt

import sockeye.constants as C
import sockeye.encoder_pt
import sockeye.transformer_pt


@pytest.mark.parametrize('dropout, factor_configs', [
    (0., None),
    (0.1, [sockeye.encoder_pt.FactorConfig(vocab_size=5,
                                           num_embed=5,
                                           combine=C.FACTORS_COMBINE_SUM,
                                           share_embedding=False)]),
])
def test_embedding_encoder(dropout, factor_configs):
    config = sockeye.encoder_pt.EmbeddingConfig(vocab_size=20, num_embed=10, dropout=dropout,
                                                factor_configs=factor_configs)
    embedding = sockeye.encoder_pt.PyTorchEmbedding(config)
    assert type(embedding) == sockeye.encoder_pt.PyTorchEmbedding


@pytest.mark.parametrize('lhuc', [
    (False,),
    (True,)
])
def test_get_transformer_encoder(lhuc):
    config = sockeye.transformer_pt.TransformerConfig(model_size=20,
                                                      attention_heads=10,
                                                      feed_forward_num_hidden=30,
                                                      act_type='test_act',
                                                      num_layers=40,
                                                      dropout_attention=1.0,
                                                      dropout_act=0.5,
                                                      dropout_prepost=0.2,
                                                      positional_embedding_type=C.LEARNED_POSITIONAL_EMBEDDING,
                                                      preprocess_sequence='test_pre',
                                                      postprocess_sequence='test_post',
                                                      max_seq_len_source=50,
                                                      max_seq_len_target=60,
                                                      use_lhuc=lhuc)
    encoder = sockeye.encoder_pt.pytorch_get_transformer_encoder(config)
    assert type(encoder) == sockeye.encoder_pt.PyTorchTransformerEncoder


def test_mx_pt_eq_transformer_encoder():
    pytest.importorskip("mxnet")
    import sockeye.transformer
    import sockeye.encoder
    import mxnet as mx
    from mxnet import np

    pt.manual_seed(13)
    mx.random.seed(13)
    config = sockeye.transformer.TransformerConfig(model_size=128,
                                                   attention_heads=8,
                                                   feed_forward_num_hidden=256,
                                                   act_type='relu',
                                                   num_layers=12,
                                                   dropout_attention=0,
                                                   dropout_act=0,
                                                   dropout_prepost=0,
                                                   positional_embedding_type=C.LEARNED_POSITIONAL_EMBEDDING,
                                                   preprocess_sequence='n',
                                                   postprocess_sequence='r',
                                                   max_seq_len_source=50,
                                                   max_seq_len_target=60,
                                                   use_lhuc=False)
    encoder_mx = sockeye.encoder.get_transformer_encoder(config, dtype=C.DTYPE_FP32)
    encoder_mx.initialize()

    encoder_pt = sockeye.encoder_pt.pytorch_get_transformer_encoder(config)
    encoder_pt.weights_from_mxnet_block(encoder_mx)

    batch = 12
    seq_len = 45
    data_mx = np.random.uniform(0, 1, (batch, seq_len, config.model_size))
    data_pt = pt.as_tensor(data_mx.asnumpy())
    lengths_mx = np.random.randint(1, seq_len, (batch,))
    lengths_pt = pt.as_tensor(lengths_mx.asnumpy())

    r1_mx, r2_mx = encoder_mx(data_mx, lengths_mx)
    r1_pt, r2_pt = encoder_pt(data_pt, lengths_pt)

    r1_mx, r2_mx = r1_mx.asnumpy(), r2_mx.asnumpy()
    r1_pt, r2_pt = r1_pt.detach().numpy(), r2_pt.detach().numpy()

    print("Max deviation:", onp.abs(r1_mx - r1_pt).max())
    assert np.allclose(r1_mx, r1_pt, atol=1e-04)
    assert np.allclose(r2_mx, r2_pt, atol=1e-04)


@pytest.mark.parametrize('vocab_size, num_embed, factor_configs, sparse',
                         [(300, 32, None, False),
                          (300, 32, None, True),
                          (300, 32, [sockeye.encoder_pt.FactorConfig(300, 8, C.FACTORS_COMBINE_CONCAT, False),
                                     sockeye.encoder_pt.FactorConfig(300, 32, C.FACTORS_COMBINE_AVERAGE, False),
                                     sockeye.encoder_pt.FactorConfig(300, 32, C.FACTORS_COMBINE_AVERAGE, True),
                                     sockeye.encoder_pt.FactorConfig(300, 32, C.FACTORS_COMBINE_SUM, True)], True)])
def test_mx_pt_eq_embedding(vocab_size, num_embed, factor_configs, sparse):
    pytest.importorskip("mxnet")
    import sockeye.encoder
    from mxnet import np

    config = sockeye.encoder.EmbeddingConfig(vocab_size=vocab_size,
                                             num_embed=num_embed,
                                             dropout=0,
                                             factor_configs=factor_configs,
                                             allow_sparse_grad=sparse)

    block_mx = sockeye.encoder.Embedding(config, None, C.DTYPE_FP32)
    block_mx.initialize()
    block_pt = sockeye.encoder_pt.PyTorchEmbedding(config, None)
    block_pt.weights_from_mxnet_block(block_mx)

    batch, seq_len, num_factors = 4, 10, len(factor_configs) + 1 if factor_configs is not None else 1
    # data_mx does not take into account different vocab sizes for factors
    data_mx = np.random.randint(0, vocab_size, (batch, seq_len, num_factors))
    data_pt = pt.as_tensor(data_mx.asnumpy())
    vl_mx = np.ones((1,))  # not used
    vl_pt = pt.as_tensor(vl_mx.asnumpy())

    r_mx, _ = block_mx(data_mx, vl_mx)
    r_pt = block_pt(data_pt)

    r_mx = r_mx.asnumpy()
    r_pt = r_pt.detach().numpy()

    assert np.allclose(r_mx, r_pt)



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.
import os
from tempfile import TemporaryDirectory

import pytest
import torch as pt

import sockeye.constants as C
import sockeye.data_io_pt
import sockeye.model_pt
import sockeye.transformer_pt


def test_mx_pt_eq_sockeye_model():
    pytest.importorskip('mxnet')
    from mxnet import np
    import sockeye.transformer
    import sockeye.encoder
    import sockeye.model

    # model setup
    source_vocab_size = target_vocab_size = 32000
    num_embed_source = num_embed_target = model_size = 512
    max_seq_len_source = max_seq_len_target = 100
    num_source_factors = 1
    num_target_factors = 1
    num_layers = 4
    weight_tying = False
    batch_size = 4
    topk_size = 200
    config_encoder = sockeye.transformer.TransformerConfig(model_size=model_size,
                                                           attention_heads=8,
                                                           feed_forward_num_hidden=256,
                                                           act_type='relu',
                                                           num_layers=num_layers,
                                                           dropout_attention=0,
                                                           dropout_act=0,
                                                           dropout_prepost=0,
                                                           positional_embedding_type=C.FIXED_POSITIONAL_EMBEDDING,
                                                           preprocess_sequence='n',
                                                           postprocess_sequence='r',
                                                           max_seq_len_source=max_seq_len_source,
                                                           max_seq_len_target=max_seq_len_target,
                                                           use_lhuc=False)
    config_encoder_pt = sockeye.transformer_pt.TransformerConfig(model_size=model_size,
                                                                 attention_heads=8,
                                                                 feed_forward_num_hidden=256,
                                                                 act_type='relu',
                                                                 num_layers=num_layers,
                                                                 dropout_attention=0,
                                                                 dropout_act=0,
                                                                 dropout_prepost=0,
                                                                 positional_embedding_type=C.FIXED_POSITIONAL_EMBEDDING,
                                                                 preprocess_sequence='n',
                                                                 postprocess_sequence='r',
                                                                 max_seq_len_source=max_seq_len_source,
                                                                 max_seq_len_target=max_seq_len_target,
                                                                 use_lhuc=False)
    config_decoder = sockeye.transformer.TransformerConfig(model_size=model_size,
                                                           attention_heads=8,
                                                           feed_forward_num_hidden=256,
                                                           act_type='relu',
                                                           num_layers=num_layers,
                                                           dropout_attention=0,
                                                           dropout_act=0,
                                                           dropout_prepost=0,
                                                           positional_embedding_type=C.FIXED_POSITIONAL_EMBEDDING,
                                                           preprocess_sequence='n',
                                                           postprocess_sequence='r',
                                                           max_seq_len_source=max_seq_len_source,
                                                           max_seq_len_target=max_seq_len_target,
                                                           depth_key_value=model_size,
                                                           use_lhuc=False)
    config_decoder_pt = sockeye.transformer_pt.TransformerConfig(model_size=model_size,
                                                                 attention_heads=8,
                                                                 feed_forward_num_hidden=256,
                                                                 act_type='relu',
                                                                 num_layers=num_layers,
                                                                 dropout_attention=0,
                                                                 dropout_act=0,
                                                                 dropout_prepost=0,
                                                                 positional_embedding_type=C.FIXED_POSITIONAL_EMBEDDING,
                                                                 preprocess_sequence='n',
                                                                 postprocess_sequence='r',
                                                                 max_seq_len_source=max_seq_len_source,
                                                                 max_seq_len_target=max_seq_len_target,
                                                                 depth_key_value=model_size,
                                                                 use_lhuc=False)
    config_embed_source = sockeye.encoder.EmbeddingConfig(vocab_size=source_vocab_size,
                                                          num_embed=num_embed_source,
                                                          dropout=0,
                                                          factor_configs=None,
                                                          allow_sparse_grad=False)
    config_embed_target = sockeye.encoder.EmbeddingConfig(vocab_size=target_vocab_size,
                                                          num_embed=num_embed_target,
                                                          dropout=0,
                                                          factor_configs=None,
                                                          allow_sparse_grad=False)
    data_statistics = sockeye.data_io_pt.DataStatistics(num_sents=0,
                                                        num_discarded=0,
                                                        num_tokens_source=0,
                                                        num_tokens_target=0,
                                                        num_unks_source=0,
                                                        num_unks_target=0,
                                                        max_observed_len_source=100,
                                                        max_observed_len_target=100,
                                                        size_vocab_source=source_vocab_size,
                                                        size_vocab_target=target_vocab_size,
                                                        length_ratio_mean=1.0,
                                                        length_ratio_std=0.001,
                                                        buckets=[],
                                                        num_sents_per_bucket=[],
                                                        average_len_target_per_bucket=[],
                                                        length_ratio_stats_per_bucket=None)
    data_config = sockeye.data_io_pt.DataConfig(data_statistics=data_statistics,
                                                max_seq_len_source=max_seq_len_source,
                                                max_seq_len_target=max_seq_len_target,
                                                num_source_factors=num_source_factors,
                                                num_target_factors=num_target_factors)
    config_length_task = None
    model_config = sockeye.model.ModelConfig(config_data=data_config,
                                             vocab_source_size=source_vocab_size,
                                             vocab_target_size=target_vocab_size,
                                             config_embed_source=config_embed_source,
                                             config_embed_target=config_embed_target,
                                             config_encoder=config_encoder,
                                             config_decoder=config_decoder,
                                             config_length_task=config_length_task,
                                             weight_tying_type=C.WEIGHT_TYING_NONE,
                                             lhuc=False,
                                             dtype=C.DTYPE_FP32)
    model_config_pt = sockeye.model.ModelConfig(config_data=data_config,
                                                vocab_source_size=source_vocab_size,
                                                vocab_target_size=target_vocab_size,
                                                config_embed_source=config_embed_source,
                                                config_embed_target=config_embed_target,
                                                config_encoder=config_encoder_pt,
                                                config_decoder=config_decoder_pt,
                                                config_length_task=config_length_task,
                                                weight_tying_type=C.WEIGHT_TYING_NONE,
                                                lhuc=False,
                                                dtype=C.DTYPE_FP32)


    # inputs
    source_inputs_mx = np.random.randint(0, max_seq_len_source, (batch_size, max_seq_len_source, num_source_factors))
    source_input_lengths_mx = np.random.randint(0, max_seq_len_source, (batch_size,))
    target_inputs_mx = np.random.randint(0, max_seq_len_target, (batch_size, max_seq_len_target, num_source_factors))
    target_input_lengths_mx = np.random.randint(0, max_seq_len_target, (batch_size,))
    source_inputs_pt = pt.tensor(source_inputs_mx.asnumpy())
    source_input_lengths_pt = pt.tensor(source_input_lengths_mx.asnumpy())
    target_inputs_pt = pt.tensor(target_inputs_mx.asnumpy())
    target_input_lengths_pt = pt.tensor(target_input_lengths_mx.asnumpy())
    step_inputs_mx = np.random.randint(0, target_vocab_size, (batch_size, num_target_factors))
    vocab_slice_ids_mx = np.random.randint(0, target_vocab_size, (topk_size,))
    step_inputs_pt = pt.tensor(step_inputs_mx.asnumpy())
    vocab_slice_ids_pt = pt.tensor(vocab_slice_ids_mx.asnumpy())

    b_mx = sockeye.model.SockeyeModel(model_config, inference_only=False, mc_dropout=False,
                                      forward_pass_cache_size=0)
    b_mx.initialize()

    b_pt = sockeye.model_pt.PyTorchSockeyeModel(model_config_pt, inference_only=False, mc_dropout=False,
                                                forward_pass_cache_size=0)

    assert b_mx.state_structure() == b_pt.state_structure()

    # test forward()
    # first run mx block to complete deferred initialization
    forward_dict_mx = b_mx(source_inputs_mx, source_input_lengths_mx, target_inputs_mx, target_input_lengths_mx)
    # get weights from mx into pt
    b_pt.weights_from_mxnet_block(b_mx)
    forward_dict_pt = b_pt(source_inputs_pt, source_input_lengths_pt, target_inputs_pt, target_input_lengths_pt)

    assert forward_dict_mx.keys() == forward_dict_pt.keys()
    logits_mx = forward_dict_mx[C.LOGITS_NAME].asnumpy()
    logits_pt = forward_dict_pt[C.LOGITS_NAME].detach().numpy()

    assert np.allclose(logits_mx, logits_pt, atol=1e-05)

    # test encode()
    source_encoded_mx, source_encoded_length_mx = b_mx.encode(source_inputs_mx, source_input_lengths_mx)
    source_encoded_pt, source_encoded_length_pt = b_pt.encode(source_inputs_pt, source_input_lengths_pt)
    assert np.allclose(source_encoded_mx.asnumpy(), source_encoded_pt.detach().numpy(), atol=1e-05)
    assert np.allclose(source_encoded_length_mx.asnumpy(), source_encoded_length_pt.detach().numpy(), atol=1e-05)

    # test encode_and_initialize()
    init_states_mx, pred_out_length_mx = b_mx.encode_and_initialize(source_inputs_mx, source_input_lengths_mx,
                                                                    constant_length_ratio=0.0)
    init_states_pt, pred_out_length_pt = b_pt.encode_and_initialize(source_inputs_pt, source_input_lengths_pt,
                                                                    constant_length_ratio=0.0)
    if config_length_task is None:
        assert np.allclose(pred_out_length_mx.asnumpy(), np.zeros_like(source_input_lengths_mx).asnumpy())
        assert np.allclose(pred_out_length_pt.detach().numpy(), pt.zeros_like(source_input_lengths_pt).detach().numpy())
    else:
        assert pred_out_length_mx.asnumpy() == pred_out_length_pt.detach().numpy()

    assert len(init_states_mx) == len(init_states_pt)
    state_structure = b_pt.decoder.state_structure()
    for s_mx, s_pt, structure in zip(init_states_mx, init_states_pt, state_structure):
        if structure != C.MASK_STATE:  # MASK state is new in Pytorch and not equivalent
            assert np.allclose(s_mx.asnumpy(), s_pt.detach().numpy(), atol=1e-05)

    # test decode_step()
    states_mx = init_states_mx
    states_pt = init_states_pt
    step_output_mx, states_mx, factor_outputs_mx = b_mx.decode_step(step_inputs_mx, states_mx,
                                                                    vocab_slice_ids=vocab_slice_ids_mx)
    step_output_pt, states_pt, factor_outputs_pt = b_pt.decode_step(step_inputs_pt, states_pt,
                                                                    vocab_slice_ids=vocab_slice_ids_pt)

    assert np.allclose(step_output_mx.asnumpy(), step_output_pt.detach().numpy(), atol=1e-05)
    assert step_output_mx.asnumpy().shape == step_output_pt.detach().numpy().shape == (batch_size, topk_size)
    assert len(factor_outputs_mx) == len(factor_outputs_pt)
    # TODO assert factor outputs equality
    assert len(states_mx) == len(states_pt)
    for s_mx, s_pt, structure in zip(states_mx, states_pt, state_structure):
        if structure != C.MASK_STATE:  # MASK state is new in Pytorch and not equivalent
            assert np.allclose(s_mx.asnumpy(), s_pt.detach().numpy(), atol=1e-05)

    from pprint import pprint
    pprint(b_mx.collect_params())
    for param_tensor in b_pt.state_dict():
        print(param_tensor, "\t", b_pt.state_dict()[param_tensor].size())

    # save & load parameters
    with TemporaryDirectory() as work_dir:
        fname = os.path.join(work_dir, 'params.pt')
        b_pt.save_parameters(fname)
        b_pt.load_parameters(fname)

    forward_dict_pt = b_pt(source_inputs_pt, source_input_lengths_pt, target_inputs_pt, target_input_lengths_pt)
    assert forward_dict_mx.keys() == forward_dict_pt.keys()
    logits_mx = forward_dict_mx[C.LOGITS_NAME].asnumpy()
    logits_pt = forward_dict_pt[C.LOGITS_NAME].detach().numpy()
    assert np.allclose(logits_mx, logits_pt, atol=1e-05)



# Copyright 2017--2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as onp
import pytest
import torch as pt

import sockeye.constants as C
import sockeye.decoder_pt
import sockeye.transformer_pt


@pytest.mark.parametrize('lhuc', [
    (False,),
    (True,)
])
def test_get_decoder(lhuc):
    config = sockeye.transformer_pt.TransformerConfig(
        model_size=20,
        attention_heads=10,
        feed_forward_num_hidden=30,
        act_type='test_act',
        num_layers=50,
        dropout_attention=0.5,
        dropout_act=0.6,
        dropout_prepost=0.1,
        positional_embedding_type=C.FIXED_POSITIONAL_EMBEDDING,
        preprocess_sequence=C.FIXED_POSITIONAL_EMBEDDING,
        postprocess_sequence='test_post_seq',
        max_seq_len_source=60,
        max_seq_len_target=70,
        use_lhuc=lhuc)
    decoder = sockeye.decoder_pt.pytorch_get_decoder(config, inference_only=False)

    assert type(decoder) == sockeye.decoder_pt.PyTorchTransformerDecoder


@pytest.mark.parametrize("inference_only", [False, True])
def test_mx_pt_eq_transformer_decoder(inference_only):
    pytest.importorskip("mxnet")
    import sockeye.transformer
    import sockeye.decoder
    import mxnet as mx
    from mxnet import np

    pt.manual_seed(13)
    mx.random.seed(13)
    config_mx = sockeye.transformer.TransformerConfig(model_size=128,
                                                      attention_heads=8,
                                                      feed_forward_num_hidden=256,
                                                      act_type='relu',
                                                      num_layers=12,
                                                      dropout_attention=0,
                                                      dropout_act=0,
                                                      dropout_prepost=0,
                                                      positional_embedding_type=C.FIXED_POSITIONAL_EMBEDDING,
                                                      preprocess_sequence='n',
                                                      postprocess_sequence='r',
                                                      max_seq_len_source=50,
                                                      max_seq_len_target=60,
                                                      depth_key_value=128,
                                                      use_lhuc=False)
    config_pt = sockeye.transformer_pt.TransformerConfig(model_size=128,
                                                         attention_heads=8,
                                                         feed_forward_num_hidden=256,
                                                         act_type='relu',
                                                         num_layers=12,
                                                         dropout_attention=0,
                                                         dropout_act=0,
                                                         dropout_prepost=0,
                                                         positional_embedding_type=C.FIXED_POSITIONAL_EMBEDDING,
                                                         preprocess_sequence='n',
                                                         postprocess_sequence='r',
                                                         max_seq_len_source=50,
                                                         max_seq_len_target=60,
                                                         depth_key_value=128,
                                                         use_lhuc=False)
    batch = 12
    encoder_seq_len = 45
    decoder_seq_len = 39 if not inference_only else 1
    encoder_outputs_mx = np.random.uniform(0, 1, (batch, encoder_seq_len, config_mx.model_size))
    encoder_outputs_pt = pt.tensor(encoder_outputs_mx.asnumpy())
    encoder_valid_length_mx = np.random.randint(1, encoder_seq_len, (batch,))
    encoder_valid_length_pt = pt.tensor(encoder_valid_length_mx.asnumpy())
    inputs_mx = np.random.uniform(0, 1, (batch, decoder_seq_len, config_mx.model_size))
    inputs_pt = pt.tensor(inputs_mx.asnumpy())

    # mx
    decoder_mx = sockeye.decoder.get_decoder(config_mx, inference_only=inference_only, dtype=C.DTYPE_FP32)
    decoder_mx.initialize()
    init_states_mx = decoder_mx.init_state_from_encoder(encoder_outputs_mx, encoder_valid_length_mx)
    output_mx, new_states_mx = decoder_mx(inputs_mx, init_states_mx)
    if inference_only:  # do a second decoder step
        output_mx, new_states_mx = decoder_mx(output_mx, new_states_mx)

    # pt
    decoder_pt = sockeye.decoder_pt.pytorch_get_decoder(config_pt, inference_only=inference_only)
    decoder_pt.weights_from_mxnet_block(decoder_mx)
    init_states_pt = decoder_pt.init_state_from_encoder(encoder_outputs_pt, encoder_valid_length_pt)
    output_pt, new_states_pt = decoder_pt(inputs_pt, init_states_pt)
    if inference_only:  # do a second decoder step
        output_pt, new_states_pt = decoder_pt(output_pt, new_states_pt)

    assert decoder_mx.state_structure() == decoder_pt.state_structure()
    assert decoder_mx.get_num_hidden() == decoder_pt.get_num_hidden()

    assert len(init_states_mx) == len(init_states_pt)
    for s_mx, s_pt, structure in zip(init_states_mx, init_states_pt, decoder_mx.state_structure()):
        if structure != C.MASK_STATE:  # MASK state is new in Pytorch and not equivalent
            assert np.allclose(s_mx.asnumpy(), s_pt.detach().numpy(), atol=1e-05)

    output_mx = output_mx.asnumpy()
    output_pt = output_pt.detach().numpy()

    print("Max deviation:", onp.abs(output_mx - output_pt).max())
    assert np.allclose(output_mx, output_pt, atol=1e-05)

    assert len(new_states_mx) == len(new_states_pt)
    for i, (s_mx, s_pt, structure) in enumerate(zip(new_states_mx, new_states_pt, decoder_mx.state_structure())):
        if structure != C.MASK_STATE:  # MASK state is new in Pytorch and not equivalent
            assert np.allclose(s_mx.asnumpy(), s_pt.detach().numpy(), atol=1e-05)



# Copyright 2017--2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import logging
from abc import abstractmethod
from dataclasses import dataclass
from typing import Optional, Union, Tuple

import mxnet as mx
from mxnet import gluon, np, npx

from . import config
from . import constants as C
from . import quantization
from . import utils

logger = logging.getLogger(__name__)


def get_activation(act_type: str) -> gluon.Block:
    """
    Returns Gluon Block for given activation type.

    Custom activation types include:
     - Swish-1, also called Sigmoid-Weighted Linear Unit (SiLU): Ramachandran et
       al. (https://arxiv.org/pdf/1710.05941.pdf), Elfwing et al.
       (https://arxiv.org/pdf/1702.03118.pdf)
     - Gaussian Error Linear Unit (GELU): Hendrycks and Gimpel
       (https://arxiv.org/pdf/1606.08415.pdf)

    :param act_type: Type of activation.
    :return: output ndarray with same shape as input.
    """
    if act_type == C.SWISH1:
        return gluon.nn.Swish()
    if act_type == C.GELU:
        return gluon.nn.GELU()
    return gluon.nn.Activation(activation=act_type)


class LHUC(gluon.HybridBlock):
    """
    Learning Hidden Unit Contribution

    David Vilar. "Learning Hidden Unit Contribution for Adapting Neural
    Machine Translation Models" NAACL 2018

    :param num_hidden: Number of hidden units of the layer to be modified.
    """
    def __init__(self,
                 num_hidden: int,
                 weight_init: Union[str, mx.init.Initializer] = mx.init.Uniform(0.1)) -> None:
        super().__init__()
        self.weight = gluon.Parameter('weight', shape=(num_hidden,), init=weight_init)

    def forward(self, data: np.ndarray) -> np.ndarray:
        # We use a sigmoid with amplitude 2 for weighting the hidden units. The
        # activation is dampened when the value of the sigmoid is close to 0, and
        # strengthened when it's close to 2 (see also original paper)
        weight = 2 * npx.activation(self.weight.data(), act_type="sigmoid")
        return weight * data


class WeightNormalization(mx.gluon.HybridBlock):
    """
    Implements Weight Normalization, see Salimans & Kingma 2016 (https://arxiv.org/abs/1602.07868).
    For a given tensor the normalization is done per hidden dimension.

    :param num_hidden: Size of the first dimension.
    :param ndim: The total number of dimensions of the weight tensor.
    """

    def __init__(self, num_hidden: int, ndim: int = 2) -> None:
        super().__init__()
        self.scale = gluon.Constant(np.ones(shape=tuple([num_hidden] + [1] * (ndim - 1))))
        self._axis_arg = tuple(range(1, ndim))

    def forward(self, weight: np.ndarray) -> np.ndarray:
        return weight / np.linalg.norm(weight, keepdims=True, axis=self._axis_arg) * self.scale.data()


class OutputLayer(gluon.HybridBlock):
    """
    Defines the output layer of Sockeye decoders. Supports weight tying and weight normalization.

    :param hidden_size: Input hidden size.
    :param vocab_size: Target vocabulary size.
    :param weight: Optional shared weight Parameter.
    :param weight_initializer: Initializer for weight.
    :param bias_initializer: Initializer for bias.
    :param dtype: Data type.
    :params params: Optional parameter dict for shared parameters.
    """

    def __init__(self,
                 hidden_size: int,
                 vocab_size: int,
                 weight: Optional[mx.gluon.Parameter] = None,
                 weight_initializer: Optional[str] = None,
                 bias_initializer: str = 'zeros',
                 dtype: str = C.DTYPE_FP32) -> None:
        super().__init__()
        self.vocab_size = vocab_size

        if dtype == C.DTYPE_INT8:
            self.scaling = gluon.Parameter('scaling',
                                           shape=(1,), init=mx.initializer.Constant(-1.0),
                                           dtype=C.DTYPE_FP32, allow_deferred_init=False)
            # This is only for inference but MXNet tries to create an
            # initializer anyway, then fails because most random
            # generators don't support int8 output.
            weight_initializer = 'zeros'
        if weight is None:
            self.weight = gluon.Parameter("weight",
                                          shape=(vocab_size, hidden_size),
                                          init=weight_initializer,
                                          dtype=dtype,
                                          allow_deferred_init=False)
        else:
            self.weight = weight

        self.bias = gluon.Parameter("bias",
                                    shape=(vocab_size,),
                                    init=bias_initializer,
                                    # Bias stays fp32 even with int8 weights.
                                    dtype=dtype if dtype != C.DTYPE_INT8 else C.DTYPE_FP32,
                                    allow_deferred_init=False)

        self._cache_key = None  # type: Optional[int]
        self._weight_slice_cache = None  # type: Optional[np.ndarray]
        self._bias_slice_cache = None  # type: Optional[np.ndarray]

    def _is_new_vocab_slices(self, x: np.ndarray) -> bool:
        # MXNet ndarrays (like Numpy ndarrays) do not support hashing, using string representation.
        x_hash = hash(str(x))
        if x_hash != self._cache_key:
            self._cache_key = x_hash
            return True
        return False

    def _take_slice(self, vocab_slice_ids: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        if self.weight.dtype == C.DTYPE_INT8:
            weight = npx.intgemm_take_weight(self.weight.data(), vocab_slice_ids)
        else:
            weight = self.weight.data().take(vocab_slice_ids, axis=0)
        bias = self.bias.data().take(vocab_slice_ids, axis=0)
        return weight, bias

    def __call__(self, data: np.ndarray, vocab_slice_ids: Optional[np.ndarray] = None) -> np.ndarray:
        if vocab_slice_ids is not None:
            # imperative, reduced matrix multiplication for vocabulary selection
            weight, bias = self._take_slice(vocab_slice_ids)
            if self._is_new_vocab_slices(vocab_slice_ids):
                weight, bias = self._take_slice(vocab_slice_ids)
                self._weight_slice_cache, self._bias_slice_cache = weight, bias
            else:
                weight, bias = self._weight_slice_cache, self._bias_slice_cache
            if self.weight.dtype == C.DTYPE_INT8:
                return npx.intgemm_fully_connected(data, weight, self.scaling.data(), bias,
                                                   num_hidden=vocab_slice_ids.shape[0],
                                                   flatten=False,
                                                   name=C.LOGITS_NAME)
            else:
                return npx.fully_connected(data,
                                           num_hidden=vocab_slice_ids.shape[0],
                                           weight=weight,
                                           bias=bias,
                                           no_bias=False,
                                           flatten=False,
                                           name=C.LOGITS_NAME)
        return super().__call__(data)

    def forward(self, data: np.ndarray) -> np.ndarray:
        if self.weight.dtype == C.DTYPE_INT8:
            return npx.intgemm_fully_connected(data,
                                               num_hidden=self.vocab_size,
                                               weight=self.weight.data(),
                                               scaling=self.scaling.data(),
                                               bias=self.bias.data(),
                                               flatten=False,
                                               name=C.LOGITS_NAME)
        else:
            return npx.fully_connected(data,
                                       num_hidden=self.vocab_size,
                                       weight=self.weight.data(),
                                       bias=self.bias.data(),
                                       no_bias=False,
                                       flatten=False,
                                       name=C.LOGITS_NAME)


@dataclass
class LengthRatioConfig(config.Config):
    num_layers: int  # Number of layers
    weight: float  # Weight of this loss


class LengthRatio(mx.gluon.HybridBlock):
    """
    Defines the length-ratio prediction layer of Sockeye.

    :param hidden_size: Encoder hidden size.
    :param num_layers: Number of layers.
    """

    def __init__(self,
                 hidden_size: int,
                 num_layers: int,
                 dtype: str = C.DTYPE_FP32) -> None:
        utils.check_condition(num_layers >= 1, "LengthRatio's num_layers has to be >=1.")
        super().__init__()
        self.num_layers = num_layers
        self.hidden_size = hidden_size

        self.layers = gluon.nn.HybridSequential()
        for l in range(num_layers - 1):
            self.layers.add(quantization.QuantizableDense(units=hidden_size, activation='tanh',
                                                          flatten=False, dtype=dtype))
        # SoftReLU activation to ensure positiveness of the predicted length ratio
        self.layers.add(quantization.QuantizableDense(units=1, activation='softrelu', flatten=False, dtype=dtype))

    def forward(self, source_encoded: np.ndarray, source_encoded_length: np.ndarray) -> np.ndarray:
        """
        Transformation to the length ratio. Returns a vector.

        :param source_encoded: Encoder representation for n elements. Shape: (n, source_encoded_length, hidden_size).
        :param source_encoded_length: A vector of encoded sequence lengths. Shape: (n,).
        :return: Predictions of the ratio length(hypothesis)/length(reference). Shape(n, 1).
        """
        # source_masked: (n, source_encoded_length, hidden_size)
        source_masked = npx.sequence_mask(source_encoded,
                                          axis=1,
                                          sequence_length=source_encoded_length,
                                          use_sequence_length=True,
                                          value=0.)
        # calculate the proper means of encoded sources
        # data: (n, hidden_size)
        data = np.sum(source_masked, axis=1, keepdims=False) / np.reshape(source_encoded_length, (-1, 1))
        # MLP. Shape: (n, 1)
        data = self.layers(data)
        # Shape: (n,)
        return np.squeeze(data)


class DotAttentionCell(gluon.HybridBlock):

    def __init__(self, dropout: float = 0.0) -> None:
        super().__init__()
        self.dropout = dropout
        self._dtype = C.DTYPE_FP32

    def cast(self, dtype):
        self._dtype = dtype
        super().cast(dtype)

    def forward(self, queries: np.ndarray, key_values: np.ndarray, heads: np.ndarray,
                lengths: Optional[np.ndarray] = None, bias: Optional[np.ndarray] = None):

        # (n*h, lq, lk)
        logits = npx.interleaved_matmul_encdec_qk(queries, key_values, heads=heads)

        if bias is not None:
            logits = logits + bias

        if lengths is not None:
            # required shape for lengths: (n*h, lq); required dtype: int32
            probs = npx.softmax(logits, axis=-1, length=lengths, use_length=True)
        else:
            probs = npx.softmax(logits, axis=-1)

        probs = npx.dropout(probs, p=self.dropout) if self.dropout > 0.0 else probs
        
        # key_values: (lk, n, dv * 2)
        # probs: (n*h, lq, lk)
        # result: (n, lq, dv)
        return npx.interleaved_matmul_encdec_valatt(key_values, probs, heads=heads)


def prepare_source_valid_lengths(valid_length: np.ndarray, query_data: np.ndarray, num_heads: int) -> np.ndarray:
    """
    Returns an int32 valid length tensor of shape (batch * num_heads, query_length) to be used in
    the softmax operation in DotAttentionCell with the length argument.
    Due to broadcast_like, dtypes of valid_length and query_data must be the same.

    :param valid_length: Valid length information. Shape: (batch,).
    :param query_data: Tensor from which the query_length dimension is derived.
                       Expected shape: (X, query_length, ...).
    :param num_heads: Number of attention heads.
    :return: int32 tensor of shape (batch * num_heads, query_length).
    """
    # (batch * heads,)
    att_valid_length = np.repeat(valid_length, repeats=num_heads, axis=0)
    att_valid_length = npx.broadcast_like(np.expand_dims(att_valid_length, axis=1),
                                          query_data,
                                          lhs_axes=(1,), rhs_axes=(1,))
    return att_valid_length.astype(dtype='int32', copy=False)


class MultiHeadAttentionBase(gluon.HybridBlock):
    """
    Base class for Multi-head attention.

    :param depth_att: Attention depth / number of hidden units.
    :param heads: Number of attention heads.
    :param depth_out: Output depth / number of output units.
    :param dropout: Dropout probability on attention scores
    :param dtype: Data type for weights
    """
    def __init__(self,
                 depth_att: int = 512,
                 heads: int = 8,
                 depth_out: int = 512,
                 dropout: float = 0.0,
                 dtype: str = C.DTYPE_FP32) -> None:
        super().__init__()
        utils.check_condition(depth_att % heads == 0,
                              "Number of heads (%d) must divide attention depth (%d)" % (heads, depth_att))
        self.depth = depth_att
        self.heads = heads
        self.depth_out = depth_out
        self.depth_per_head = self.depth // self.heads

        self.dot_att = DotAttentionCell(dropout=dropout)
        self.ff_out = quantization.QuantizableDense(in_units=depth_att, units=depth_out,
                                                    flatten=False, use_bias=False, dtype=dtype)

    def _attend(self,
                queries: np.ndarray,
                key_values: np.ndarray,
                lengths: Optional[np.ndarray] = None,
                bias: Optional[np.ndarray] = None) -> np.ndarray:
        """
        Returns context vectors of multi-head dot attention.

        :param queries: Query tensor. Shape: (query_max_length, batch_size, depth).
        :param key_values: Keys. Shape: (memory_max_length, batch_size, depth * 2).
        :param lengths: Optional lengths of keys. Shape: (batch_size*heads,).
        :param bias: Optional 3d bias.
        :return: Context vectors. Shape: (batch_size, query_max_length, output_depth).
        """

        # (query_max_length, batch, depth)
        contexts = self.dot_att(queries, key_values, self.heads, lengths, bias)

        # (query_max_length, batch, output_depth)
        contexts = self.ff_out(contexts)

        return contexts


class AutoregressiveLayer(mx.gluon.HybridBlock):
    @property
    @abstractmethod
    def num_state_tensors(self) -> int:
        """ Number of state tensors returned by the layer """
        raise NotImplementedError

    @property
    @abstractmethod
    def needs_mask(self) -> bool:
        """ Whether the layer makes use of a mask tensor or not """
        raise NotImplementedError

    @abstractmethod
    def get_state_shape(self, batch_size) -> Tuple:
        """
        :param batch_size: current batch size
        :return: dimensions of each output state (assuming all of them have the same shape)
        """
        raise NotImplementedError

    @abstractmethod
    def forward(self, inputs: np.ndarray, previous_states: np.ndarray, *args) -> Tuple:
        """
        :param inputs: layer input
        :param previous_states: Previous states array or list of arrays
        :param args: layer-specific arguments and/or arguments to be ignored
        :return: layer output and new states
        """
        raise NotImplementedError


class MultiHeadSelfAttention(MultiHeadAttentionBase, AutoregressiveLayer):
    """
    Multi-head self-attention. Independent linear projections of inputs serve as
    queries, keys, and values for the attention.

    :param depth_att: Attention depth / number of hidden units.
    :param heads: Number of attention heads.
    :param depth_out: Output depth / number of output units.
    :param dropout: Dropout probability on attention scores
    :param dtype: Data type for weights
    """
    def __init__(self,
                 depth_att: int = 512,
                 heads: int = 8,
                 depth_out: int = 512,
                 dropout: float = 0.0,
                 dtype: str = C.DTYPE_FP32) -> None:
        super().__init__(depth_att, heads, depth_out, dropout, dtype)

        self.depth_att = depth_att
        self.ff_in = quantization.QuantizableDense(in_units=depth_att, units=depth_att * 3,
                                                   flatten=False, use_bias=False, dtype=dtype)

    @property
    def num_state_tensors(self) -> int:
        """ Number of state tensors returned by the layer """
        return 1

    @property
    def needs_mask(self) -> bool:
        """ Whether the layer makes use of a mask tensor or not """
        return True

    def get_state_shape(self, batch_size: int) -> Tuple:
        """
        :param batch_size: current batch size
        :return: dimensions of each output state (assuming all of them have the same shape)
        """
        # shape: (length, batch, key_depth + value_depth)
        return 0, batch_size, self.depth_out * 2

    def forward(self,
                inputs: np.ndarray,
                previous_states: Optional[np.ndarray] = None,
                input_lengths: Optional[np.ndarray] = None,
                bias: Optional[np.ndarray] = None,
                *args) -> Tuple[np.ndarray, np.ndarray]:  # mypy: ignore
        """
        Computes multi-head attention on a set of inputs, serving as queries, keys, and values.
        If sequence lengths are provided, they will be used to mask the attention scores.
        A bias mask may also be used to mask the attention scores.
        May also use a cache of previously computed inputs.
        Returns a ndarray of shape (batch, max_length, output_depth).

        :param inputs: Input Data. Shape: (max_length, batch, input_depth).
        :param input_lengths: Optional lengths of inputs to mask attention scores. Shape: (batch, 1).
        :param bias: Optional 3d bias tensor to mask attention scores.
        :param previous_states: Optional list with two ndarrays - previous input's keys and values.
                                Shape: 2 * (batch, max_length+1, depth_att).
        :return: ndarray of shape (batch, max_length, output_depth).
        """

        proj = self.ff_in(inputs)
        queries, kv_1, kv_2 = np.split(proj, 3, axis=2)
        states = np.concatenate((kv_1, kv_2), axis=2)

        if previous_states is not None:
            states = np.concatenate((previous_states, states), axis=0)

        return self._attend(queries, states, lengths=input_lengths, bias=bias), states


class MultiHeadAttention(MultiHeadAttentionBase):
    """
    Multi-head attention layer for queries independent from keys/values.

    :param depth_att: Attention depth / number of hidden units.
    :param heads: Number of attention heads.
    :param depth_out: Output depth / number of output units.
    :param depth_key_value: Dimension of input key and value vectors.
    :param dropout: Dropout probability on attention scores
    :param dtype: Data type for weights
    """

    def __init__(self,
                 depth_att: int = 512,
                 heads: int = 8,
                 depth_out: int = 512,
                 dropout: float = 0.0,
                 dtype: str = C.DTYPE_FP32,
                 depth_key_value: int = 0) -> None:
        super().__init__(depth_att, heads, depth_out, dropout, dtype)

        self.ff_q = quantization.QuantizableDense(in_units=depth_out, units=depth_att,
                                                  flatten=False, use_bias=False, dtype=dtype)
        self.ff_kv = quantization.QuantizableDense(in_units=depth_key_value, units=2 * depth_att,
                                                   flatten=False, use_bias=False, dtype=dtype)

    def forward(self, queries: np.ndarray,
                memory: np.ndarray,
                memory_lengths: Optional[np.ndarray] = None,
                bias: Optional[np.ndarray] = None,
                projected_memory_kv: Optional[np.ndarray] = None) -> np.ndarray:  # mypy: ignore
        """
        Computes multi-head attention for queries given a memory tensor.
        If sequence lengths are provided, they will be used to mask the attention scores.
        A bias mask may also be used to mask the attention scores.
        Returns an ndarray of shape (max_length, batch, output_depth).

        :param queries: Query tensor. Shape: (query_max_length, batch, input_depth).
        :param memory: Memory data to attend to. Shape: (memory_max_length, batch, input_depth).
        :param memory_lengths: Optional lengths of memory to mask attention scores. Shape: (batch, 1).
        :param bias: Optional 3d bias tensor to mask attention scores.
        :param projected_memory_kv: Optional previously projected memory keys and values.
        :return: ndarray of shape (query_seq_len, batch, output_depth).
        """

        queries = self.ff_q(queries)
        kv = projected_memory_kv if projected_memory_kv is not None else self.ff_kv(memory)
        return self._attend(queries, kv, bias=bias, lengths=memory_lengths)


# unused, not ported
class PlainDotAttention(gluon.HybridBlock):
    """
    Dot attention layer for queries independent from keys/values.
    """
    def __init__(self):
        super().__init__()
        self.dot_att = DotAttentionCell()

    def forward(self, queries: np.ndarray, memory: np.ndarray, memory_lengths: np.ndarray) -> np.ndarray:
        """
        Returns an ndarray of shape (batch, max_length, output_depth).

        :param queries: ndarray of shape (queries_max_length, batch, input_depth).
        :param memory: ndarray of shape (memory_max_length, batch, input_depth).
        :param memory_lengths: ndarray of shape (batch, 1).
        :return: ndarray of shape (queries_max_length, batch, output_depth).
       """

        # (queries_max_length, batch, output_depth)
        return self.dot_att(queries, memory, 1, memory_lengths, None)


# unused, not ported
class ProjectedDotAttention(gluon.HybridBlock):
    """
    Dot attention layer for queries independent from keys/values.

    :param num_hidden: Attention depth / number of hidden units.
    """

    def __init__(self,
                 num_hidden: int,
                 dtype: str) -> None:
        super().__init__()
        self.num_hidden = num_hidden
        self.q2h = quantization.QuantizableDense(units=num_hidden, flatten=False, use_bias=True, dtype=dtype)
        self.kv2h = quantization.QuantizableDense(units=num_hidden * 2, flatten=False, use_bias=True, dtype=dtype)
        self.dot_att = DotAttentionCell()

    def forward(self, queries: np.ndarray,
                      memory: np.ndarray,
                      memory_lengths: np.ndarray) -> np.ndarray:
        """
        Apply project, apply dot attention and return new context vectors.

        :param queries: ndarray of shape (queries_max_length, batch, input_num_hidden).
        :param memory: ndarray of shape (memory_max_length, batch, input_num_hidden).
        :param memory_lengths: ndarray of shape (batch, 1).
        :return: ndarray of shape (queries_max_length, batch, num_hidden).
        """
        # (memory_max_length, batch, num_hidden * 2)
        combined = self.kv2h(memory)

        # (queries_max_length, batch, num_hidden)
        queries = self.q2h(queries)

        # (queries_max_length, batch, num_hidden)
        contexts = self.dot_att(queries, combined, 1, memory_lengths, None)

        return contexts


def get_positional_embeddings(length, depth) -> np.ndarray:
    utils.check_condition(depth % 2 == 0, "Positional embeddings require an even embedding size it "
                                          "is however %d." % depth)
    # (1, depth)
    channels = np.arange(depth // 2).reshape((1, -1))

    # (length, 1)
    positions = np.arange(0, length).reshape((-1, 1))
    scaled_positions = positions / np.power(10000, (2 * channels) / depth)
    # sinusoids:
    sin = np.sin(scaled_positions)
    # cosines:
    cos = np.cos(scaled_positions)
    # interleave: (length, num_embed)
    encodings = np.hstack([sin, cos])
    return encodings


class PositionalEmbeddings(gluon.HybridBlock):
    """
    Takes an encoded sequence and adds sinusoidal or learned positional embeddings as in Vaswani et al, 2017 to it.

    :param weight_type: type of embeddings, fixed or learned.
    :param num_embed: Embedding size.
    :param max_seq_len: Maximum sequence length.
    :param scale_up_input: If True, scales input data up by num_embed ** 0.5.
    :param scale_down_positions: If True, scales positional embeddings down by num_embed ** -0.5.
    :param weight_init: Optional initializer for learned embeddings.
    """

    def __init__(self,
                 weight_type: str,
                 num_embed: int,
                 max_seq_len: int,
                 scale_up_input: bool,
                 scale_down_positions: bool,
                 weight_init: Optional[Union[str, mx.init.Initializer]] = None) -> None:
        utils.check_condition(num_embed % 2 == 0, "Positional embeddings require an even embedding size it "
                                                  "is however %d." % num_embed)
        super().__init__()
        self.weight_type = weight_type
        self.num_embed = num_embed
        self.max_seq_len = max_seq_len
        self.scale_up_input = scale_up_input
        self.scale_down_positions = scale_down_positions

        if self.weight_type == C.FIXED_POSITIONAL_EMBEDDING:
            pos_weight = get_positional_embeddings(length=self.max_seq_len, depth=self.num_embed)
            if self.scale_down_positions:
                pos_weight *= self.num_embed ** -0.5
            self.weight = mx.gluon.Constant(pos_weight)
        elif self.weight_type == C.LEARNED_POSITIONAL_EMBEDDING:
            self.weight = mx.gluon.Parameter('weight', shape=(self.max_seq_len, self.num_embed), init=weight_init)
        else:
            raise ValueError("weight_type '%s' is not supported!" % self.weight_type)

    def forward(self, data, steps):  # pylint: disable=arguments-differ
        """
        Applies positional embeddings to input data.

        :param data: Input data. Shape: (batch, length or 1, num_embed)
        :param steps: Optional steps input. If given, shape is (batch_size or 1, seq_len,)

        :return: Data with positional embeddings added
        """
        # (length, num_embed)
        if steps is None:
            # (batch, length, num_embed)
            pos_embedding = npx.slice_like(np.expand_dims(self.weight.data(), axis=0), data, axes=(1,))
        else:
            # (batch_size or 1, seq_len, num_embed)
            pos_embedding = npx.embedding(steps, self.weight.data(), self.max_seq_len, self.num_embed)

        if self.weight_type == 'fixed':
            pos_embedding = npx.stop_gradient(pos_embedding)

        if self.scale_up_input:
            data = data * (self.num_embed ** 0.5)

        return data + pos_embedding


class SSRU(AutoregressiveLayer):
    """
    Simpler Simple Recurrent Unit

    Kim et al, "From Research to Production and Back: Ludicrously Fast Neural Machine Translation" WNGT 2019

    Variant of an LSTM cell aimed at reducing computational dependency across time steps.
    Formally described as:

    (1) f[t] = sigmoid(W1[t] * x[t] + b[t])
    (2) c[t] = f[t] . c[t-1] + (1 - f[t]) . W2[t] * x[t]
    (3) h = ReLU(c[t])

    where:
        . represents elementwise multiplication;
        x[t] is the input at time step t;
        f[t] is the output of the forget gate at time step t;
        c[t] is the cell state at time step t;
        h is the output of the unit.

    :param model_size: number of hidden units
    :param inference_only: flag used to indicate execution at inference time
    :param dtype: data type
    """
    def __init__(self,
                 model_size: int,
                 inference_only: bool,
                 dtype: str = C.DTYPE_FP32) -> None:
        super(SSRU, self).__init__()

        self.model_size = model_size
        self.inference_only = inference_only

        self.cell_state_transform = self._inference_cell_state_transform \
                                    if inference_only else self._training_cell_state_transform

        self.forget_gate = quantization.QuantizableDense(in_units=model_size,
                                                         units=model_size,
                                                         activation="sigmoid",
                                                         flatten=False,
                                                         dtype=dtype)

        self.linear = quantization.QuantizableDense(in_units=model_size,
                                                    units=model_size,
                                                    use_bias=False,
                                                    flatten=False,
                                                    dtype=dtype)

    @property
    def num_state_tensors(self) -> int:
        """ Number of state tensors returned by the layer """
        return 1

    @property
    def needs_mask(self) -> bool:
        """ Whether the layer makes use of a mask tensor or not """
        return False

    def get_state_shape(self, batch_size: int) -> Tuple:
        """
        :param batch_size: current batch size
        :return: dimensions of each output state (assuming all of them have the same shape)
        """
        return 1, batch_size, self.model_size

    @staticmethod
    def _training_cell_state_transform(previous_cell_state, weighted_inputs, forget_rates) -> Tuple[np.ndarray,
                                                                                                    np.ndarray]:
        """Update SSRU cell at training time"""

        def _time_step_update(step_input_and_forget_rate, previous_step_state) -> Tuple[np.ndarray, np.ndarray]:
            """
            Recurrently update the SSRU cell state for one time step.

            :param step_input_and_forget_rate: List = [step_input, forget_rate]
            :param previous_step_state: cell state at (t-1)
            :return: twice the current time step state. NOTE: The first instance will be stacked in the final
            foreach output and the second will be the input to the next time_step_update iteration.
            """
            step_input, forget_rate = step_input_and_forget_rate  # each of shape (batch_size, model_size)
            current_step_state = forget_rate * previous_step_state + step_input
            return current_step_state, current_step_state

        # (max_length, batch, input_depth), (batch, input_depth)
        cell_state, last_step_state = npx.foreach(_time_step_update,
                                                  [weighted_inputs, forget_rates],
                                                  np.squeeze(previous_cell_state, axis=0))

        return cell_state, np.expand_dims(last_step_state, axis=0)

    @staticmethod
    def _inference_cell_state_transform(previous_cell_state, weighted_inputs, forget_rates) -> Tuple[np.ndarray,
                                                                                                     np.ndarray]:
        """Update SSRU cell at inference time"""
        new_step_state = forget_rates * previous_cell_state + weighted_inputs  # (1, batch, input_depth)
        return new_step_state, new_step_state

    def forward(self, inputs: np.ndarray, previous_states: np.ndarray, *args) -> Tuple[np.ndarray, np.ndarray]:
        """
        :param inputs: input data. Shape: (max_length, batch, input_depth).
        :param previous_states: previous cell states. Shape: (max_length, batch, input_depth)
        :return: cell output and new cell states.  Both with shape (max_length, batch, input_depth).
        """
        forget_rates = self.forget_gate(inputs)
        weighted_inputs = (1 - forget_rates) * self.linear(inputs)

        cell_state, last_step_state = self.cell_state_transform(previous_states, weighted_inputs, forget_rates)

        return npx.relu(cell_state), last_step_state



# Copyright 2017--2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
import random
from tempfile import TemporaryDirectory
from typing import Optional, List, Tuple

import pytest

from sockeye import constants as C
from sockeye import vocab
from sockeye.utils import SockeyeError, get_tokens, seed_rngs

seed_rngs(12)

define_bucket_tests = [(50, 10, [10, 20, 30, 40, 50]),
                       (50, 20, [20, 40, 50]),
                       (50, 50, [50]),
                       (5, 10, [5]),
                       (11, 5, [5, 10, 11]),
                       (19, 10, [10, 19])]


@pytest.mark.parametrize("max_seq_len, step, expected_buckets", define_bucket_tests)
def test_define_buckets(max_seq_len, step, expected_buckets):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    buckets = data_io.define_buckets(max_seq_len, step=step)
    assert buckets == expected_buckets


define_parallel_bucket_tests = [(50, 50, 10, True, 1.0, [(10, 10), (20, 20), (30, 30), (40, 40), (50, 50)]),
                                (50, 50, 10, True, 0.5,
                                 [(10, 5), (20, 10), (30, 15), (40, 20), (50, 25), (50, 30), (50, 35), (50, 40),
                                  (50, 45), (50, 50)]),
                                (10, 10, 10, True, 0.1,
                                 [(10, 2), (10, 3), (10, 4), (10, 5), (10, 6), (10, 7), (10, 8), (10, 9), (10, 10)]),
                                (10, 5, 10, True, 0.01, [(10, 2), (10, 3), (10, 4), (10, 5)]),
                                (50, 50, 10, True, 2.0,
                                 [(5, 10), (10, 20), (15, 30), (20, 40), (25, 50), (30, 50), (35, 50), (40, 50),
                                  (45, 50), (50, 50)]),
                                (5, 10, 10, True, 10.0, [(2, 10), (3, 10), (4, 10), (5, 10)]),
                                (5, 10, 10, True, 11.0, [(2, 10), (3, 10), (4, 10), (5, 10)]),
                                (50, 50, 50, True, 0.5, [(50, 25), (50, 50)]),
                                (50, 50, 50, True, 1.5, [(33, 50), (50, 50)]),
                                (75, 75, 50, True, 1.5, [(33, 50), (66, 75), (75, 75)]),
                                (50, 50, 8, False, 1.5, [(8, 8), (16, 16), (24, 24), (32, 32), (40, 40), (48, 48),
                                                         (50, 50)]),
                                (50, 75, 8, False, 1.5, [(8, 8), (16, 16), (24, 24), (32, 32), (40, 40), (48, 48),
                                                         (50, 56), (50, 64), (50, 72), (50, 75)])]


@pytest.mark.parametrize("max_seq_len_source, max_seq_len_target, bucket_width, bucket_scaling, length_ratio,"
                         "expected_buckets", define_parallel_bucket_tests)
def test_define_parallel_buckets(max_seq_len_source, max_seq_len_target, bucket_width, bucket_scaling, length_ratio,
                                 expected_buckets):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    buckets = data_io.define_parallel_buckets(max_seq_len_source, max_seq_len_target, bucket_width=bucket_width,
                                              bucket_scaling=bucket_scaling, length_ratio=length_ratio)
    assert buckets == expected_buckets


get_bucket_tests = [([10, 20, 30, 40, 50], 50, 50),
                    ([10, 20, 30, 40, 50], 11, 20),
                    ([10, 20, 30, 40, 50], 9, 10),
                    ([10, 20, 30, 40, 50], 51, None),
                    ([10, 20, 30, 40, 50], 1, 10),
                    ([10, 20, 30, 40, 50], 0, 10),
                    ([], 50, None)]


@pytest.mark.parametrize("buckets, length, expected_bucket",
                         get_bucket_tests)
def test_get_bucket(buckets, length, expected_bucket):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    bucket = data_io.get_bucket(length, buckets)
    assert bucket == expected_bucket


tokens2ids_tests = [(["a", "b", "c"], {"a": 1, "b": 0, "c": 300, C.UNK_SYMBOL: 12}, [1, 0, 300]),
                    (["a", "x", "c"], {"a": 1, "b": 0, "c": 300, C.UNK_SYMBOL: 12}, [1, 12, 300])]


@pytest.mark.parametrize("tokens, vocab, expected_ids", tokens2ids_tests)
def test_tokens2ids(tokens, vocab, expected_ids):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    ids = data_io.tokens2ids(tokens, vocab)
    assert ids == expected_ids


@pytest.mark.parametrize("tokens, expected_ids", [(["1", "2", "3", "0"], [1, 2, 3, 0]), ([], [])])
def test_strids2ids(tokens, expected_ids):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    ids = data_io.strids2ids(tokens)
    assert ids == expected_ids


sequence_reader_tests = [(["1 2 3", "2", "", "2 2 2"], False, False, False),
                         (["a b c", "c"], True, False, False),
                         (["a b c", ""], True, False, False),
                         (["a b c", "c"], True, True, True)]


@pytest.mark.parametrize("sequences, use_vocab, add_bos, add_eos", sequence_reader_tests)
def test_sequence_reader(sequences, use_vocab, add_bos, add_eos):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    with TemporaryDirectory() as work_dir:
        path = os.path.join(work_dir, 'input')
        with open(path, 'w') as f:
            for sequence in sequences:
                print(sequence, file=f)

        vocabulary = vocab.build_vocab(sequences) if use_vocab else None

        reader = data_io.SequenceReader(path, vocabulary=vocabulary, add_bos=add_bos, add_eos=add_eos)

        read_sequences = [s for s in reader]
        assert len(read_sequences) == len(sequences)

        if vocabulary is None:
            with pytest.raises(SockeyeError) as e:
                data_io.SequenceReader(path, vocabulary=vocabulary, add_bos=True)
            assert str(e.value) == "Adding a BOS or EOS symbol requires a vocabulary"

            expected_sequences = [data_io.strids2ids(get_tokens(s)) if s else None for s in sequences]
            assert read_sequences == expected_sequences
        else:
            expected_sequences = [data_io.tokens2ids(get_tokens(s), vocabulary) if s else None for s in sequences]
            if add_bos:
                expected_sequences = [[vocabulary[C.BOS_SYMBOL]] + s if s else None for s in expected_sequences]
            if add_eos:
                expected_sequences = [s + [vocabulary[C.EOS_SYMBOL]] if s else None for s in expected_sequences]
            assert read_sequences == expected_sequences


@pytest.mark.parametrize("source_iterables, target_iterables",
                         [
                             (
                                     [[[0], [1, 1], [2], [3, 3, 3]], [[0], [1, 1], [2], [3, 3, 3]]],
                                     [[[0], [1]]]
                             ),
                             (
                                     [[[0], [1, 1]], [[0], [1, 1]]],
                                     [[[0], [1, 1], [2], [3, 3, 3]]]
                             ),
                             (
                                     [[[0], [1, 1]]],
                                     [[[0], [1, 1], [2], [3, 3, 3]]]
                             ),
                         ])
def test_nonparallel_iter(source_iterables, target_iterables):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    with pytest.raises(SockeyeError) as e:
        list(data_io.parallel_iter(source_iterables, target_iterables))
    assert str(e.value) == "Different number of lines in source(s) and target(s) iterables."


@pytest.mark.parametrize("source_iterables, target_iterables",
                         [
                             (
                                     [[[0], [1, 1]], [[0], [1]]],
                                     [[[0], [1]]]
                             )
                         ])
def test_not_source_token_parallel_iter(source_iterables, target_iterables):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    with pytest.raises(SockeyeError) as e:
        list(data_io.parallel_iter(source_iterables, target_iterables))
    assert str(e.value).startswith("Source sequences are not token-parallel")


@pytest.mark.parametrize("source_iterables, target_iterables",
                         [
                             (
                                     [[[0], [1]]],
                                     [[[0], [1, 1]], [[0], [1]]],
                             )
                         ])
def test_not_target_token_parallel_iter(source_iterables, target_iterables):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    with pytest.raises(SockeyeError) as e:
        list(data_io.parallel_iter(source_iterables, target_iterables))
    assert str(e.value).startswith("Target sequences are not token-parallel")


@pytest.mark.parametrize("source_iterables, target_iterables, expected",
                         [
                             (
                                     [[[0], [1, 1]], [[0], [1, 1]]],
                                     [[[0], [1]]],
                                     [([[0], [0]], [[0]]), ([[1, 1], [1, 1]], [[1]])]
                             ),
                             (
                                     [[[0], None], [[0], None]],
                                     [[[0], [1]]],
                                     [([[0], [0]], [[0]])]
                             ),
                             (
                                     [[[0], [1, 1]], [[0], [1, 1]]],
                                     [[[0], None]],
                                     [([[0], [0]], [[0]])]
                             ),
                             (
                                     [[None, [1, 1]], [None, [1, 1]]],
                                     [[None, [1]]],
                                     [([[1, 1], [1, 1]], [[1]])]
                             ),
                             (
                                     [[None, [1]]],
                                     [[None, [1, 1]], [None, [1, 1]]],
                                     [([[1]], [[1, 1], [1, 1]])]
                             ),
                             (
                                     [[None, [1, 1]], [None, [1, 1]]],
                                     [[None, None]],
                                     []
                             )
                         ])
def test_parallel_iter(source_iterables, target_iterables, expected):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    assert list(data_io.parallel_iter(source_iterables, target_iterables)) == expected


def test_sample_based_define_bucket_batch_sizes():
    pytest.importorskip('mxnet')
    from sockeye import data_io
    batch_type = C.BATCH_TYPE_SENTENCE
    batch_size = 32
    max_seq_len = 100
    buckets = data_io.define_parallel_buckets(max_seq_len, max_seq_len, 10, 1, 1.5)
    bucket_batch_sizes = data_io.define_bucket_batch_sizes(buckets=buckets,
                                                           batch_size=batch_size,
                                                           batch_type=batch_type,
                                                           batch_num_devices=1,
                                                           data_target_average_len=[None] * len(buckets))
    for bbs in bucket_batch_sizes:
        assert bbs.batch_size == batch_size
        assert bbs.average_target_words_per_batch == bbs.bucket[1] * batch_size


@pytest.mark.parametrize("batch_num_devices,length_ratio,batch_sentences_multiple_of,expected_batch_sizes", [
    # Reference batch sizes manually inspected for sanity.  Note that for
    # very unbalanced lengths, the last batch can be very large.  This is
    # due to the requirement for any size batch (total elements) to fit into
    # the same allocated space for MXNet's memory sharing.
    (1, 0.5, 1, [200.0, 100.0, 67.0, 50.0, 40.0, 33.0, 29.0, 25.0, 22.0, 41.0]),
    (2, 0.5, 1, [200.0, 100.0, 66.0, 50.0, 40.0, 34.0, 28.0, 24.0, 22.0, 40.0]),
    (8, 0.5, 1, [200.0, 96.0, 64.0, 48.0, 40.0, 32.0, 32.0, 24.0, 24.0, 40.0]),
    (1, 1.5, 1, [100.0, 50.0, 33.0, 25.0, 20.0, 20.0, 20.0, 20.0]),
    (1, 1.5, 8, [96.0, 48.0, 32.0, 24.0, 16.0, 16.0, 16.0, 24.0])])
def test_word_based_define_bucket_batch_sizes(batch_num_devices, length_ratio, batch_sentences_multiple_of,
                                              expected_batch_sizes):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    batch_type = C.BATCH_TYPE_WORD
    batch_size = 1000
    max_seq_len = 50
    buckets = data_io.define_parallel_buckets(max_seq_len, max_seq_len, 10, True, length_ratio)
    bucket_batch_sizes = data_io.define_bucket_batch_sizes(buckets=buckets,
                                                           batch_size=batch_size,
                                                           batch_type=batch_type,
                                                           batch_num_devices=batch_num_devices,
                                                           data_target_average_len=[None] * len(buckets),
                                                           batch_sentences_multiple_of=batch_sentences_multiple_of)
    max_num_words = 0
    # last bucket batch size is different
    for bbs, expected_batch_size in zip(bucket_batch_sizes, expected_batch_sizes):
        assert bbs.batch_size == expected_batch_size
        expected_average_target_words_per_batch = expected_batch_size * bbs.bucket[1]
        assert bbs.average_target_words_per_batch == expected_average_target_words_per_batch
        max_num_words = max(max_num_words, bbs.batch_size * max(*bbs.bucket))

    last_bbs = bucket_batch_sizes[-1]
    min_expected_batch_size = round((batch_size / last_bbs.bucket[1]) / batch_num_devices)
    assert last_bbs.batch_size >= min_expected_batch_size
    last_bbs_num_words = last_bbs.batch_size * max(*last_bbs.bucket)
    assert last_bbs_num_words >= max_num_words


@pytest.mark.parametrize("batch_num_devices,length_ratio,batch_sentences_multiple_of,expected_batch_sizes", [
    # Reference batch sizes manually inspected for sanity.
    (1, 0.5, 1, [200, 100, 66, 50, 40, 33, 28, 25, 22, 20]),
    (2, 0.5, 1, [200, 100, 66, 50, 40, 32, 28, 24, 22, 20]),
    (8, 0.5, 1, [200, 96, 64, 48, 40, 32, 24, 24, 16, 16]),
    (1, 1.5, 1, [100, 50, 33, 25, 20, 20, 20, 20]),
    (1, 1.5, 8, [96, 48, 32, 24, 16, 16, 16, 16])])
def test_max_word_based_define_bucket_batch_sizes(batch_num_devices, length_ratio, batch_sentences_multiple_of,
                                                  expected_batch_sizes):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    batch_type = C.BATCH_TYPE_MAX_WORD
    batch_size = 1000
    max_seq_len = 50
    buckets = data_io.define_parallel_buckets(max_seq_len, max_seq_len, 10, True, length_ratio)
    bucket_batch_sizes = data_io.define_bucket_batch_sizes(buckets=buckets,
                                                           batch_size=batch_size,
                                                           batch_type=batch_type,
                                                           batch_num_devices=batch_num_devices,
                                                           data_target_average_len=[None] * len(buckets),
                                                           batch_sentences_multiple_of=batch_sentences_multiple_of)
    for bbs, expected_batch_size in zip(bucket_batch_sizes, expected_batch_sizes):
        assert bbs.batch_size == expected_batch_size
        expected_average_target_words_per_batch = expected_batch_size * bbs.bucket[1]
        assert bbs.average_target_words_per_batch == expected_average_target_words_per_batch


def _get_random_bucketed_data(buckets: List[Tuple[int, int]],
                              min_count: int,
                              max_count: int,
                              bucket_counts: Optional[List[Optional[int]]] = None):
    """
    Get random bucket data.

    :param buckets: The list of buckets.
    :param min_count: The minimum number of samples that will be sampled if no exact count is given.
    :param max_count: The maximum number of samples that will be sampled if no exact count is given.
    :param bucket_counts: For each bucket an optional exact example count can be given. If it is not given it will be
                         sampled.
    :return: The random source, target and label arrays.
    """
    pytest.importorskip('mxnet')
    from mxnet import np
    if bucket_counts is None:
        bucket_counts = [None for _ in buckets]
    bucket_counts = [random.randint(min_count, max_count) if given_count is None else given_count
                     for given_count in bucket_counts]
    source = [np.array(np.random.randint(0, 10, (count, random.randint(1, bucket[0]), 1))) for count, bucket in
              zip(bucket_counts, buckets)]
    target = [np.array(np.random.randint(0, 10, (count, random.randint(2, bucket[1]), 1))) for count, bucket in
              zip(bucket_counts, buckets)]
    return source, target


def test_parallel_data_set():
    pytest.importorskip('mxnet')
    from sockeye import data_io
    from mxnet import np
    buckets = data_io.define_parallel_buckets(100, 100, 10, 1, 1.0)
    source, target = _get_random_bucketed_data(buckets, min_count=0, max_count=5)

    def check_equal(arrays1, arrays2):
        assert len(arrays1) == len(arrays2)
        for a1, a2 in zip(arrays1, arrays2):
            assert np.array_equal(a1, a2)

    with TemporaryDirectory() as work_dir:
        dataset = data_io.ParallelDataSet(source, target)
        fname = os.path.join(work_dir, 'dataset')
        dataset.save(fname)
        dataset_loaded = data_io.ParallelDataSet.load(fname)
        check_equal(dataset.source, dataset_loaded.source)
        check_equal(dataset.target, dataset_loaded.target)


def test_parallel_data_set_fill_up():
    pytest.importorskip('mxnet')
    from sockeye import data_io
    batch_size = 32
    buckets = data_io.define_parallel_buckets(100, 100, 10, 1, 1.0)
    bucket_batch_sizes = data_io.define_bucket_batch_sizes(buckets,
                                                           batch_size,
                                                           batch_type=C.BATCH_TYPE_SENTENCE,
                                                           batch_num_devices=1,
                                                           data_target_average_len=[None] * len(buckets))
    dataset = data_io.ParallelDataSet(*_get_random_bucketed_data(buckets, min_count=1, max_count=5))

    dataset_filled_up = dataset.fill_up(bucket_batch_sizes)
    assert len(dataset_filled_up.source) == len(dataset.source)
    assert len(dataset_filled_up.target) == len(dataset.target)
    for bidx in range(len(dataset)):
        bucket_batch_size = bucket_batch_sizes[bidx].batch_size
        assert dataset_filled_up.source[bidx].shape[0] == bucket_batch_size
        assert dataset_filled_up.target[bidx].shape[0] == bucket_batch_size


def test_get_permutations():
    pytest.importorskip('mxnet')
    from sockeye import data_io
    from mxnet import np
    data = [list(range(3)), list(range(1)), list(range(7)), []]
    bucket_counts = [len(d) for d in data]

    permutation, inverse_permutation = data_io.get_permutations(bucket_counts)
    assert len(permutation) == len(inverse_permutation) == len(bucket_counts) == len(data)

    for d, p, pi in zip(data, permutation, inverse_permutation):
        p = p.astype(np.int32)
        pi = pi.astype(np.int32)
        p_set = set(p.tolist())
        pi_set = set(pi.tolist())
        assert len(p_set) == len(p)
        assert len(pi_set) == len(pi)
        assert p_set - pi_set == set()
        if d:
            d = np.array(d)
            assert (d[p][pi] == d).all()
        else:
            assert len(p_set) == 1


def test_parallel_data_set_permute():
    pytest.importorskip('mxnet')
    from sockeye import data_io
    batch_size = 5
    buckets = data_io.define_parallel_buckets(100, 100, 10, True, 1.0)
    bucket_batch_sizes = data_io.define_bucket_batch_sizes(buckets,
                                                           batch_size,
                                                           batch_type=C.BATCH_TYPE_SENTENCE,
                                                           batch_num_devices=1,
                                                           data_target_average_len=[None] * len(buckets))
    dataset = data_io.ParallelDataSet(*_get_random_bucketed_data(buckets, min_count=0, max_count=5)).fill_up(
        bucket_batch_sizes)

    permutations, inverse_permutations = data_io.get_permutations(dataset.get_bucket_counts())

    assert len(permutations) == len(inverse_permutations) == len(dataset)
    dataset_restored = dataset.permute(permutations).permute(inverse_permutations)
    assert len(dataset) == len(dataset_restored)
    for buck_idx in range(len(dataset)):
        num_samples = dataset.source[buck_idx].shape[0]
        if num_samples:
            assert (dataset.source[buck_idx] == dataset_restored.source[buck_idx]).all()
            assert (dataset.target[buck_idx] == dataset_restored.target[buck_idx]).all()
        else:
            assert not dataset_restored.source[buck_idx]
            assert not dataset_restored.target[buck_idx]


def test_get_batch_indices():
    pytest.importorskip('mxnet')
    from sockeye import data_io
    max_bucket_size = 50
    batch_size = 10
    buckets = data_io.define_parallel_buckets(100, 100, 10, True, 1.0)
    bucket_batch_sizes = data_io.define_bucket_batch_sizes(buckets,
                                                           batch_size,
                                                           batch_type=C.BATCH_TYPE_SENTENCE,
                                                           batch_num_devices=1,
                                                           data_target_average_len=[None] * len(buckets))
    dataset = data_io.ParallelDataSet(*_get_random_bucketed_data(buckets=buckets,
                                                                 min_count=1,
                                                                 max_count=max_bucket_size))

    indices = data_io.get_batch_indices(dataset, bucket_batch_sizes=bucket_batch_sizes)

    # check for valid indices
    for buck_idx, start_pos in indices:
        assert 0 <= buck_idx < len(dataset)
        assert 0 <= start_pos < len(dataset.source[buck_idx]) - batch_size + 1

    # check that all indices are used for a filled-up dataset
    dataset = dataset.fill_up(bucket_batch_sizes)
    indices = data_io.get_batch_indices(dataset, bucket_batch_sizes=bucket_batch_sizes)
    all_bucket_indices = set(list(range(len(dataset))))
    computed_bucket_indices = set([i for i, j in indices])

    assert not all_bucket_indices - computed_bucket_indices


@pytest.mark.parametrize("buckets, expected_default_bucket_key",
                         [([(10, 10), (20, 20), (30, 30), (40, 40), (50, 50)], (50, 50)),
                          ([(5, 10), (10, 20), (15, 30), (25, 50), (20, 40)], (25, 50))])
def test_get_default_bucket_key(buckets, expected_default_bucket_key):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    default_bucket_key = data_io.get_default_bucket_key(buckets)
    assert default_bucket_key == expected_default_bucket_key


get_parallel_bucket_tests = [([(10, 10), (20, 20), (30, 30), (40, 40), (50, 50)], 50, 50, 4, (50, 50)),
                             ([(10, 10), (20, 20), (30, 30), (40, 40), (50, 50)], 50, 10, 4, (50, 50)),
                             ([(10, 10), (20, 20), (30, 30), (40, 40), (50, 50)], 20, 10, 1, (20, 20)),
                             ([(10, 10)], 20, 10, None, None),
                             ([], 20, 10, None, None),
                             ([(10, 11)], 11, 10, None, None),
                             ([(11, 10)], 11, 10, 0, (11, 10))]


@pytest.mark.parametrize("buckets, source_length, target_length, expected_bucket_index, expected_bucket",
                         get_parallel_bucket_tests)
def test_get_parallel_bucket(buckets, source_length, target_length, expected_bucket_index, expected_bucket):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    bucket_index, bucket = data_io.get_parallel_bucket(buckets, source_length, target_length)
    assert bucket_index == expected_bucket_index
    assert bucket == expected_bucket


@pytest.mark.parametrize("sources, targets, expected_num_sents, expected_mean, expected_std",
                         [([[[1, 1, 1], [2, 2, 2], [3, 3, 3]]],
                           [[[1, 1, 1], [2, 2, 2], [3, 3, 3]]], 3, 1.0, 0.0),
                          ([[[1, 1], [2, 2], [3, 3]]],
                           [[[1, 1, 1], [2, 2, 2], [3, 3, 3]]], 3, 1.5, 0.0),
                          ([[[1, 1, 1], [2, 2], [3, 3, 3, 3, 3, 3, 3]]],
                           [[[1, 1, 1], [2], [3, 3, 3]]], 2, 0.75, 0.25)])
def test_calculate_length_statistics(sources, targets, expected_num_sents, expected_mean, expected_std):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    from mxnet import np
    length_statistics = data_io.calculate_length_statistics(sources, targets, 5, 5)
    assert len(sources[0]) == len(targets[0])
    assert length_statistics.num_sents == expected_num_sents
    assert np.isclose(length_statistics.length_ratio_mean, expected_mean)
    assert np.isclose(length_statistics.length_ratio_std, expected_std)


@pytest.mark.parametrize("sources, targets",
                         [
                             ([[[1, 1, 1], [2, 2, 2], [3, 3, 3]],
                               [[1, 1, 1], [2, 2], [3, 3, 3]]],
                              [[[1, 1, 1], [2, 2, 2], [3, 3, 3]]])
                         ])
def test_non_parallel_calculate_length_statistics(sources, targets):
    pytest.importorskip('mxnet')
    from sockeye import data_io
    with pytest.raises(SockeyeError):
        data_io.calculate_length_statistics(sources, targets, 5, 5)


def test_get_training_data_iters():
    pytest.importorskip('mxnet')
    from sockeye import data_io
    from mxnet import np
    from sockeye.test_utils import tmp_digits_dataset
    train_line_count = 100
    train_line_count_empty = 0
    train_max_length = 30
    dev_line_count = 20
    dev_max_length = 30
    expected_mean = 1.0
    expected_std = 0.0
    test_line_count = 20
    test_line_count_empty = 0
    test_max_length = 30
    batch_size = 5
    num_source_factors = num_target_factors = 1
    with tmp_digits_dataset("tmp_corpus",
                            train_line_count, train_line_count_empty, train_max_length - C.SPACE_FOR_XOS,
                            dev_line_count, dev_max_length - C.SPACE_FOR_XOS,
                            test_line_count, test_line_count_empty,
                            test_max_length - C.SPACE_FOR_XOS) as data:
        # tmp common vocab
        vcb = vocab.build_from_paths([data['train_source'], data['train_target']])

        train_iter, val_iter, config_data, data_info = data_io.get_training_data_iters(
            sources=[data['train_source']],
            targets=[data['train_target']],
            validation_sources=[data['dev_source']],
            validation_targets=[data['dev_target']],
            source_vocabs=[vcb],
            target_vocabs=[vcb],
            source_vocab_paths=[None],
            target_vocab_paths=[None],
            shared_vocab=True,
            batch_size=batch_size,
            batch_type=C.BATCH_TYPE_SENTENCE,
            batch_num_devices=1,
            max_seq_len_source=train_max_length,
            max_seq_len_target=train_max_length,
            bucketing=True,
            bucket_width=10)
        assert isinstance(train_iter, data_io.ParallelSampleIter)
        assert isinstance(val_iter, data_io.ParallelSampleIter)
        assert isinstance(config_data, data_io.DataConfig)
        assert data_info.sources == [data['train_source']]
        assert data_info.targets == [data['train_target']]
        assert data_info.source_vocabs == [None]
        assert data_info.target_vocabs == [None]
        assert config_data.data_statistics.max_observed_len_source == train_max_length
        assert config_data.data_statistics.max_observed_len_target == train_max_length
        assert np.isclose(config_data.data_statistics.length_ratio_mean, expected_mean)
        assert np.isclose(config_data.data_statistics.length_ratio_std, expected_std)

        assert train_iter.batch_size == batch_size
        assert val_iter.batch_size == batch_size
        assert train_iter.default_bucket_key == (train_max_length, train_max_length)
        assert val_iter.default_bucket_key == (dev_max_length, dev_max_length)
        assert train_iter.dtype == 'float32'

        # test some batches
        bos_id = vcb[C.BOS_SYMBOL]
        eos_id = vcb[C.EOS_SYMBOL]
        expected_first_target_symbols = np.full((batch_size, 1), bos_id, dtype='float32')
        for epoch in range(2):
            while train_iter.iter_next():
                batch = train_iter.next()
                assert isinstance(batch, data_io.Batch)
                source = batch.source
                target = batch.target
                label = batch.labels[C.TARGET_LABEL_NAME]  # TODO: still 2-shape: (batch, length)
                length_ratio_label = batch.labels[C.LENRATIO_LABEL_NAME]
                assert source.shape[0] == target.shape[0] == label.shape[0] == batch_size
                assert source.shape[2] == target.shape[2] == num_source_factors == num_target_factors
                # target first symbol should be BOS
                # each source sequence contains one EOS symbol
                assert np.sum(source == eos_id) == batch_size
                assert np.array_equal(target[:, 0], expected_first_target_symbols)
                # label first symbol should be 2nd target symbol
                assert np.array_equal(label[:, 0], target[:, 1, 0])
                # each label sequence contains one EOS symbol
                assert np.sum(label == eos_id) == batch_size
            train_iter.reset()


def _data_batches_equal(db1, db2) -> bool:
    from mxnet import np
    equal = True
    equal = equal and np.allclose(db1.source, db2.source)
    equal = equal and np.allclose(db1.source_length, db2.source_length)
    equal = equal and np.allclose(db1.target, db2.target)
    equal = equal and np.allclose(db1.target_length, db2.target_length)
    equal = equal and db1.labels.keys() == db2.labels.keys()
    equal = equal and db1.samples == db2.samples
    equal = equal and db1.tokens == db2.tokens
    return equal


def test_parallel_sample_iter():
    pytest.importorskip('mxnet')
    from sockeye import data_io

    batch_size = 2
    buckets = data_io.define_parallel_buckets(100, 100, 10, True, 1.0)
    # The first bucket is going to be empty:
    bucket_counts = [0] + [None] * (len(buckets) - 1)
    bucket_batch_sizes = data_io.define_bucket_batch_sizes(buckets,
                                                           batch_size,
                                                           batch_type=C.BATCH_TYPE_SENTENCE,
                                                           batch_num_devices=1,
                                                           data_target_average_len=[None] * len(buckets))

    dataset = data_io.ParallelDataSet(*_get_random_bucketed_data(buckets, min_count=0, max_count=5,
                                                                 bucket_counts=bucket_counts))
    it = data_io.ParallelSampleIter(dataset, buckets, batch_size, bucket_batch_sizes)

    with TemporaryDirectory() as work_dir:
        # Test 1
        it.next()
        expected_batch = it.next()

        fname = os.path.join(work_dir, "saved_iter")
        it.save_state(fname)

        it_loaded = data_io.ParallelSampleIter(dataset, buckets, batch_size, bucket_batch_sizes)
        it_loaded.reset()
        it_loaded.load_state(fname)
        loaded_batch = it_loaded.next()
        assert _data_batches_equal(expected_batch, loaded_batch)

        # Test 2
        it.reset()
        expected_batch = it.next()
        it.save_state(fname)

        it_loaded = data_io.ParallelSampleIter(dataset, buckets, batch_size, bucket_batch_sizes)
        it_loaded.reset()
        it_loaded.load_state(fname)

        loaded_batch = it_loaded.next()
        assert _data_batches_equal(expected_batch, loaded_batch)

        # Test 3
        it.reset()
        expected_batch = it.next()
        it.save_state(fname)
        it_loaded = data_io.ParallelSampleIter(dataset, buckets, batch_size, bucket_batch_sizes)
        it_loaded.reset()
        it_loaded.load_state(fname)

        loaded_batch = it_loaded.next()
        assert _data_batches_equal(expected_batch, loaded_batch)

        while it.iter_next():
            it.next()
            it_loaded.next()
        assert not it_loaded.iter_next()


def test_sharded_parallel_sample_iter():
    pytest.importorskip('mxnet')
    from sockeye import data_io
    batch_size = 2
    buckets = data_io.define_parallel_buckets(100, 100, 10, 1, 1.0)
    # The first bucket is going to be empty:
    bucket_counts = [0] + [None] * (len(buckets) - 1)
    bucket_batch_sizes = data_io.define_bucket_batch_sizes(buckets,
                                                           batch_size,
                                                           batch_type=C.BATCH_TYPE_SENTENCE,
                                                           batch_num_devices=1,
                                                           data_target_average_len=[None] * len(buckets))

    dataset1 = data_io.ParallelDataSet(*_get_random_bucketed_data(buckets, min_count=0, max_count=5,
                                                                  bucket_counts=bucket_counts))
    dataset2 = data_io.ParallelDataSet(*_get_random_bucketed_data(buckets, min_count=0, max_count=5,
                                                                  bucket_counts=bucket_counts))

    with TemporaryDirectory() as work_dir:
        shard1_fname = os.path.join(work_dir, 'shard1')
        shard2_fname = os.path.join(work_dir, 'shard2')
        dataset1.save(shard1_fname)
        dataset2.save(shard2_fname)
        shard_fnames = [shard1_fname, shard2_fname]

        it = data_io.ShardedParallelSampleIter(shard_fnames, buckets, batch_size, bucket_batch_sizes)

        # Test 1
        it.next()
        expected_batch = it.next()

        fname = os.path.join(work_dir, "saved_iter")
        it.save_state(fname)

        it_loaded = data_io.ShardedParallelSampleIter(shard_fnames, buckets, batch_size, bucket_batch_sizes)
        it_loaded.reset()
        it_loaded.load_state(fname)
        loaded_batch = it_loaded.next()
        assert _data_batches_equal(expected_batch, loaded_batch)

        # Test 2
        it.reset()
        expected_batch = it.next()
        it.save_state(fname)

        it_loaded = data_io.ShardedParallelSampleIter(shard_fnames, buckets, batch_size, bucket_batch_sizes)
        it_loaded.reset()
        it_loaded.load_state(fname)

        loaded_batch = it_loaded.next()
        assert _data_batches_equal(expected_batch, loaded_batch)

        # Test 3
        it.reset()
        expected_batch = it.next()
        it.save_state(fname)
        it_loaded = data_io.ShardedParallelSampleIter(shard_fnames, buckets, batch_size, bucket_batch_sizes)
        it_loaded.reset()
        it_loaded.load_state(fname)

        loaded_batch = it_loaded.next()
        assert _data_batches_equal(expected_batch, loaded_batch)

        while it.iter_next():
            it.next()
            it_loaded.next()
        assert not it_loaded.iter_next()


def test_sharded_parallel_sample_iter_num_batches():
    pytest.importorskip('mxnet')
    from sockeye import data_io
    num_shards = 2
    batch_size = 2
    num_batches_per_bucket = 10
    buckets = data_io.define_parallel_buckets(100, 100, 10, 1, 1.0)
    bucket_counts = [batch_size * num_batches_per_bucket for _ in buckets]
    num_batches_per_shard = num_batches_per_bucket * len(buckets)
    num_batches = num_shards * num_batches_per_shard
    bucket_batch_sizes = data_io.define_bucket_batch_sizes(buckets,
                                                           batch_size,
                                                           batch_type=C.BATCH_TYPE_SENTENCE,
                                                           batch_num_devices=1,
                                                           data_target_average_len=[None] * len(buckets))

    dataset1 = data_io.ParallelDataSet(*_get_random_bucketed_data(buckets, min_count=0, max_count=5,
                                                                  bucket_counts=bucket_counts))
    dataset2 = data_io.ParallelDataSet(*_get_random_bucketed_data(buckets, min_count=0, max_count=5,
                                                                  bucket_counts=bucket_counts))
    with TemporaryDirectory() as work_dir:
        shard1_fname = os.path.join(work_dir, 'shard1')
        shard2_fname = os.path.join(work_dir, 'shard2')
        dataset1.save(shard1_fname)
        dataset2.save(shard2_fname)
        shard_fnames = [shard1_fname, shard2_fname]

        it = data_io.ShardedParallelSampleIter(shard_fnames, buckets, batch_size, bucket_batch_sizes)

        num_batches_seen = 0
        while it.iter_next():
            it.next()
            num_batches_seen += 1
        assert num_batches_seen == num_batches


def test_sharded_and_parallel_iter_same_num_batches():
    """ Tests that a sharded data iterator with just a single shard produces as many shards as an iterator directly
    using the same dataset. """
    pytest.importorskip('mxnet')
    from sockeye import data_io
    batch_size = 2
    num_batches_per_bucket = 10
    buckets = data_io.define_parallel_buckets(100, 100, 10, 1, 1.0)
    bucket_counts = [batch_size * num_batches_per_bucket for _ in buckets]
    num_batches = num_batches_per_bucket * len(buckets)
    bucket_batch_sizes = data_io.define_bucket_batch_sizes(buckets,
                                                           batch_size,
                                                           batch_type=C.BATCH_TYPE_SENTENCE,
                                                           batch_num_devices=1,
                                                           data_target_average_len=[None] * len(buckets))

    dataset = data_io.ParallelDataSet(*_get_random_bucketed_data(buckets, min_count=0, max_count=5,
                                                                 bucket_counts=bucket_counts))

    with TemporaryDirectory() as work_dir:
        shard_fname = os.path.join(work_dir, 'shard1')
        dataset.save(shard_fname)
        shard_fnames = [shard_fname]

        it_sharded = data_io.ShardedParallelSampleIter(shard_fnames, buckets, batch_size, bucket_batch_sizes)

        it_parallel = data_io.ParallelSampleIter(dataset, buckets, batch_size, bucket_batch_sizes)

        num_batches_seen = 0
        while it_parallel.iter_next():
            assert it_sharded.iter_next()
            it_parallel.next()
            it_sharded.next()
            num_batches_seen += 1
        assert num_batches_seen == num_batches

        print("Resetting...")
        it_sharded.reset()
        it_parallel.reset()

        num_batches_seen = 0
        while it_parallel.iter_next():
            assert it_sharded.iter_next()
            it_parallel.next()
            it_sharded.next()

            num_batches_seen += 1

        assert num_batches_seen == num_batches


def test_create_target_and_shifted_label_sequences():
    pytest.importorskip('mxnet')
    from sockeye import data_io
    from mxnet import np
    target_and_label = np.array([[C.BOS_ID, 4, 17, 35, 12, C.EOS_ID, C.PAD_ID, C.PAD_ID],
                                 [C.BOS_ID, 15, 23, 23, 77, 55, 22, C.EOS_ID],
                                 [C.BOS_ID, 4, C.EOS_ID, C.PAD_ID, C.PAD_ID, C.PAD_ID, C.PAD_ID, C.PAD_ID]])
    target_and_label = np.expand_dims(target_and_label, axis=2)
    expected_lengths = np.array([5, 7, 2])

    target, label = data_io.create_target_and_shifted_label_sequences(target_and_label)

    assert target.shape[0] == label.shape[0] == target_and_label.shape[0]
    assert target.shape[1] == label.shape[1] == target_and_label.shape[1] - 1
    lengths = (target != C.PAD_ID).sum(axis=1).squeeze()
    assert np.allclose(lengths, expected_lengths)



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# Called before importing mxnet or any module that imports mxnet
import shutil

from . import initial_setup
initial_setup.handle_env_cli_arg()

import logging
from typing import Optional

from . import arguments
from sockeye.log import setup_main_logger
import torch as pt
from .model_pt import make_pytorch_model_from_mxnet_model
from . import constants as C

import os
import shutil

logger = logging.getLogger(__name__)


def convert(model: str, checkpoint: Optional[int], skip_backup=False):
    setup_main_logger(file_logging=False)
    import sockeye.model
    logger.info(f"############### Loading MXNet model from '{model}', checkpoint={checkpoint}")
    model_mx, _, _ = sockeye.model.load_model(model,
                                              checkpoint=checkpoint,
                                              hybridize=False,
                                              inference_only=False)
    logger.info(f"############### Constructing PyTorch model...")
    model_pt = make_pytorch_model_from_mxnet_model(model_mx)
    if checkpoint is None:
        params_fname = os.path.join(model, C.PARAMS_BEST_NAME)
    else:
        params_fname = os.path.join(model, C.PARAMS_NAME % checkpoint)

    if not skip_backup:
        mx_params_fname = params_fname + ".mx"
        logger.info(f"############### Backing up MXNet parameter file to '{mx_params_fname}'...")
        shutil.copyfile(params_fname, mx_params_fname)
    else:
        logger.info(f"############### Overwriting MXNet parameter file at '{params_fname}'...")
    logger.info(f"############### Saving PyTorch parameters to '{params_fname}...'")
    model_pt.save_parameters(params_fname)


def main():
    params = arguments.ConfigArgumentParser(description='PyTorch converter CLI')
    params.add_argument("--model", "-m", required=True,
                        help="Model directory containing trained model.")
    params.add_argument('--checkpoint', '-c',
                        default=None,
                        type=int,
                        nargs='+',
                        help='Checkpoint index to load (int). If not given, chooses best checkpoints for model(s). ')
    params.add_argument('--skip-backup',
                        action='store_true',
                        help='Do not backup MXNet parameter file to <param fname>.mx. Default: %(default)s.')
    args = params.parse_args()

    convert(args.model, args.checkpoint, args.skip_backup)


if __name__ == '__main__':
    main()



# Copyright 2017--2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as onp
import pytest
import torch as pt

import sockeye.layers_pt


def test_lhuc():
    num_hidden = 50
    batch_size = 10
    inp = pt.rand(batch_size, num_hidden)

    lhuc = sockeye.layers_pt.PyTorchLHUC(num_hidden=num_hidden)
    pt.nn.init.zeros_(lhuc.weight)
    out = lhuc(inp)
    pt.testing.assert_allclose(inp, out)

    lhuc = sockeye.layers_pt.PyTorchLHUC(num_hidden=num_hidden)
    pt.nn.init.constant_(lhuc.weight, 20.0)
    out = lhuc(inp)
    pt.testing.assert_allclose(2 * inp, out)


def test_mx_pt_eq_lhuc():
    mxnet = pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.layers

    num_hidden = 50
    batch_size = 10
    inp_mx = np.random.uniform(0, 1, size=(batch_size, num_hidden))
    inp_pt = pt.as_tensor(inp_mx.asnumpy())
    b_mx = sockeye.layers.LHUC(num_hidden=num_hidden, weight_init='zeros')
    b_mx.initialize()
    b_pt = sockeye.layers_pt.PyTorchLHUC(num_hidden=num_hidden)
    pt.nn.init.zeros_(b_pt.weight)

    out_mx = b_mx(inp_mx).asnumpy()
    out_pt = b_pt(inp_pt).detach().numpy()

    assert onp.allclose(out_mx, out_pt)

    b_mx = sockeye.layers.LHUC(num_hidden=num_hidden, weight_init=mxnet.init.Constant(value=20.0))
    b_mx.initialize()
    b_pt = sockeye.layers_pt.PyTorchLHUC(num_hidden=num_hidden)
    pt.nn.init.constant_(b_pt.weight, 20.0)

    out_mx = b_mx(inp_mx).asnumpy()
    out_pt = b_pt(inp_pt).detach().numpy()

    assert onp.allclose(out_mx, out_pt)


def test_weight_normalization():
    expected_norm = onp.array([1., 1.])
    weight = pt.tensor([[1., 2.],
                        [3., 4.]])
    weight_norm = sockeye.layers_pt.PyTorchWeightNormalization(num_hidden=2)
    norm_weight = weight_norm(weight).detach().numpy()
    assert onp.allclose(onp.linalg.norm(norm_weight, axis=1), expected_norm)


def test_mx_pt_eq_weight_normalization():
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.layers

    num_hidden = 3
    weight_mx = np.random.uniform(0, 1, size=(num_hidden, 4))
    weight_pt = pt.as_tensor(weight_mx.asnumpy())
    b_mx = sockeye.layers.WeightNormalization(num_hidden=num_hidden)
    b_mx.initialize()
    b_pt = sockeye.layers_pt.PyTorchWeightNormalization(num_hidden=num_hidden)

    result_mx = b_mx(weight_mx).asnumpy()
    result_pt = b_pt(weight_pt).detach().numpy()

    assert np.allclose(result_mx, result_pt)


def test_positional_embeddings():
    num_embed = 32
    max_seq_len = 10
    scale_up_input = False
    scale_down_positions = False
    data_len = 5
    data = pt.zeros(2, data_len, num_embed)

    # fixed embeddings
    expected_fixed_embedding = sockeye.layers_pt.pytorch_get_positional_embeddings(data_len, num_embed)
    b = sockeye.layers_pt.PyTorchPositionalEmbeddings(weight_type='fixed',
                                                      num_embed=num_embed,
                                                      max_seq_len=max_seq_len,
                                                      scale_up_input=scale_up_input,
                                                      scale_down_positions=scale_down_positions)
    # no steps
    out = b(data, None)
    pt.testing.assert_allclose(out[0], expected_fixed_embedding)
    pt.testing.assert_allclose(out[1], expected_fixed_embedding)

    # steps
    steps = pt.tensor([2, 3, 1, 1, 1]).unsqueeze(0)
    out = b(data, steps)
    pt.testing.assert_allclose(out[0, 0], expected_fixed_embedding[2])
    pt.testing.assert_allclose(out[1, 0], expected_fixed_embedding[2])
    pt.testing.assert_allclose(out[0, 1], expected_fixed_embedding[3])
    pt.testing.assert_allclose(out[1, 1], expected_fixed_embedding[3])
    pt.testing.assert_allclose(out[0, 2], expected_fixed_embedding[1])
    pt.testing.assert_allclose(out[1, 2], expected_fixed_embedding[1])

    # learned embeddings
    b = sockeye.layers_pt.PyTorchPositionalEmbeddings(weight_type='learned',
                                                      num_embed=num_embed,
                                                      max_seq_len=max_seq_len,
                                                      scale_up_input=scale_up_input,
                                                      scale_down_positions=scale_down_positions)
    pt.nn.init.constant_(b.weight, val=1.0)
    expected_learned_embeddings = pt.ones(data_len, num_embed)
    out = b(data, None)
    pt.testing.assert_allclose(out[0], expected_learned_embeddings)


@pytest.mark.parametrize('data_len, num_embed, scale_up_input, scale_down_positions, steps',
                         [(5, 32, False, False, None),
                          (5, 32, False, False, [2, 3]),
                          (10, 32, True, False, [1, 3]),
                          (4, 32, False, True, [2, 3])])
def test_mx_pt_eq_positional_embeddings(data_len, num_embed, scale_up_input, scale_down_positions, steps):
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.layers

    max_seq_len = 10
    data_mx = np.random.uniform(0, 1, (2, data_len, num_embed))
    data_pt = pt.as_tensor(data_mx.asnumpy())
    if steps is None:
        steps_mx, steps_pt = None, None
    else:
        steps_mx = np.array(steps).reshape((-1, 1))
        steps_pt = pt.as_tensor(steps).unsqueeze(1)

    b_mx = sockeye.layers.PositionalEmbeddings(weight_type='fixed',
                                               num_embed=num_embed,
                                               max_seq_len=max_seq_len,
                                               scale_up_input=scale_up_input,
                                               scale_down_positions=scale_down_positions,
                                               weight_init=None)
    b_mx.initialize()
    r_mx = b_mx(data_mx, steps_mx).asnumpy()

    b_pt = sockeye.layers_pt.PyTorchPositionalEmbeddings(weight_type='fixed',
                                                         num_embed=num_embed,
                                                         max_seq_len=max_seq_len,
                                                         scale_up_input=scale_up_input,
                                                         scale_down_positions=scale_down_positions)
    b_pt.weights_from_mxnet_block(b_mx)
    r_pt = b_pt(data_pt, steps_pt).detach().numpy()

    np.allclose(r_mx, r_pt)


def test_mx_pt_eq_get_positional_embeddings():
    pytest.importorskip("mxnet")
    import sockeye.layers

    data_len = 5
    num_embed = 32

    embed_mx = sockeye.layers.get_positional_embeddings(data_len, num_embed).asnumpy()
    embed_pt = sockeye.layers_pt.pytorch_get_positional_embeddings(data_len, num_embed).detach().numpy()

    assert onp.allclose(embed_mx, embed_pt)


def test_output_layer():
    num_hidden = 32
    vocab_size = 64
    data = pt.ones(2, 10, num_hidden)
    vocab_slice_ids = pt.tensor([4, 7, 23])

    b = sockeye.layers_pt.PyTorchOutputLayer(num_hidden, vocab_size)
    assert b.weight.data.shape == (vocab_size, num_hidden)

    output = b(data, None)
    assert output.shape == (2, 10, vocab_size)
    reduced_output = output.index_select(-1, vocab_slice_ids)

    output_restricted = b(data, vocab_slice_ids)
    assert output_restricted.shape == (2, 10, len(vocab_slice_ids))

    pt.testing.assert_allclose(output_restricted, reduced_output)


def test_mx_pt_eq_output_layer():
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.layers

    num_hidden = 32
    vocab_size = 64
    data_mx = np.random.uniform(0, 1, (2, 10, num_hidden))
    data_pt = pt.as_tensor(data_mx.asnumpy())
    vocab_slice_ids_mx = np.array([4, 7, 23])
    vocab_slice_ids_pt = pt.tensor([4, 7, 23])

    b_mx = sockeye.layers.OutputLayer(num_hidden, vocab_size)
    b_mx.initialize()

    b_pt = sockeye.layers_pt.PyTorchOutputLayer(num_hidden, vocab_size)
    b_pt.weights_from_mxnet_block(b_mx)
    assert b_pt.weight.size() == (vocab_size, num_hidden)

    out_mx = b_mx(data_mx, None)
    assert out_mx.shape == (2, 10, vocab_size)

    out_pt = b_pt(data_pt, None)
    assert out_pt.shape == (2, 10, vocab_size)

    assert np.allclose(out_mx.asnumpy(), out_pt.detach().numpy(), atol=1e-06)

    reduced_out_mx = out_mx.take(vocab_slice_ids_mx, axis=-1).asnumpy()
    reduced_out_pt = pt.index_select(out_pt, 2, vocab_slice_ids_pt).detach().numpy()
    assert np.allclose(reduced_out_mx, reduced_out_pt, atol=1e-06)

    out_restricted_mx = b_mx(data_mx, vocab_slice_ids_mx).asnumpy()
    out_restricted_pt = b_pt(data_pt, vocab_slice_ids_pt).detach().numpy()
    assert out_restricted_mx.shape == (2, 10, len(vocab_slice_ids_mx))
    assert out_restricted_pt.shape == (2, 10, len(vocab_slice_ids_pt))

    assert onp.allclose(out_restricted_mx, out_restricted_pt, atol=1e-06)


@pytest.mark.parametrize('qlen, kvlen, batch_size',
                         [(10, 9, 1), (1, 1, 1), (3, 32, 128)])
def test_mx_pt_eq_interleaved_matmul_encdec_qk(qlen, kvlen, batch_size):
    pytest.importorskip("mxnet")
    from mxnet import np, npx
    import sockeye.layers

    hidden = 32
    q_mx = np.random.uniform(0, 1, (qlen, batch_size, hidden))
    kv_mx = np.random.uniform(0, 1, (kvlen, batch_size, hidden * 2))
    heads = 4
    q_pt = pt.as_tensor(q_mx.asnumpy())
    kv_pt = pt.as_tensor(kv_mx.asnumpy())

    assert np.allclose(q_pt.numpy(), q_mx.asnumpy())
    assert np.allclose(kv_pt.numpy(), kv_mx.asnumpy())

    r0 = npx.interleaved_matmul_encdec_qk(q_mx, kv_mx, heads=heads).asnumpy()
    r1 = sockeye.layers_pt.pytorch_interleaved_matmul_encdec_qk(q_pt, kv_pt, heads=heads).detach().numpy()
    assert np.allclose(r0, r1)


@pytest.mark.parametrize('qlen, kvlen, batch_size',
                         [(10, 9, 1), (1, 1, 1), (3, 32, 128)])
def test_mx_pt_eq_interleaved_matmul_encdec_valatt(qlen, kvlen, batch_size):
    pytest.importorskip("mxnet")
    from mxnet import np, npx
    import sockeye.layers

    hidden = 32
    kv_mx = np.random.uniform(0, 1, (kvlen, batch_size, hidden * 2))
    heads = 4
    kv_pt = pt.as_tensor(kv_mx.asnumpy())
    att = np.random.uniform(0, 1, (batch_size * heads, qlen, kvlen))
    attpt = pt.as_tensor(att.asnumpy())
    r0 = npx.interleaved_matmul_encdec_valatt(kv_mx, att, heads=heads).asnumpy()
    r1 = sockeye.layers_pt.pytorch_interleaved_matmul_encdec_valatt(kv_pt, attpt, heads=heads).numpy()
    assert np.allclose(r0, r1)


@pytest.mark.parametrize('qlen, kvlen, batch_size, hidden, heads',
                         [
                             (10, 9, 1, 32, 8),
                             (1, 1, 1, 4, 1),
                             (13, 13, 20, 8, 2),
                             (3, 32, 64, 128, 2)
                         ])
def test_mx_pt_eq_dot_attention_cell(qlen, kvlen, batch_size, hidden, heads):
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.layers
    import sockeye.transformer
    import sockeye.transformer_pt

    q_mx = np.random.uniform(0, 1, (qlen, batch_size, hidden))
    kv_mx = np.random.uniform(0, 1, (kvlen, batch_size, hidden * 2))
    q_pt = pt.as_tensor(q_mx.asnumpy())
    kv_pt = pt.as_tensor(kv_mx.asnumpy())

    if qlen == kvlen:  # self-attention case
        bias_mx = sockeye.transformer.AutoRegressiveBias()
        bias_mx.initialize()
        autoregr_mx = bias_mx(q_mx.transpose(1, 0, 2))
        mx_args = (None, autoregr_mx)  # no source mask, autoregr mask

        mask_pt = sockeye.transformer_pt.AutoRegressiveMask()
        att_mask_pt = mask_pt(q_pt.permute(1, 0, 2))

    else:  # cross-attention
        lengths_mx = np.random.randint(1, kvlen, (batch_size,), )
        valid_lengths_mx = sockeye.layers.prepare_source_valid_lengths(lengths_mx, q_mx.transpose(1, 0, 2), heads)
        mx_args = (valid_lengths_mx, None)  # source mask, no autoregr mask

        lengths_pt = pt.tensor(lengths_mx.asnumpy())
        att_mask_pt = sockeye.layers_pt.prepare_source_length_mask(lengths_pt, heads, kvlen)

    b_mx = sockeye.layers.DotAttentionCell(dropout=0.0)
    b_mx.initialize()
    b_pt = sockeye.layers_pt.PyTorchDotAttentionCell(dropout=0.0, heads=heads)

    r_mx = b_mx(q_mx, kv_mx, heads, *mx_args).asnumpy()
    r_pt = b_pt(q_pt, kv_pt, mask=att_mask_pt).detach().numpy()

    assert np.allclose(r_mx, r_pt, atol=1e-06)


@pytest.mark.parametrize('qlen, kvlen, batch_size, hidden, heads',
                         [(10, 9, 1, 12, 4), (1, 1, 1, 4, 1), (3, 32, 15, 64, 2)])
def test_mx_pt_eq_multi_head_attention_base(qlen, kvlen, batch_size, hidden, heads):
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.layers

    q_mx = np.random.uniform(0, 1, (qlen, batch_size, hidden))
    kv_mx = np.random.uniform(0, 1, (kvlen, batch_size, hidden * 2))
    q_pt = pt.as_tensor(q_mx.asnumpy())
    kv_pt = pt.as_tensor(kv_mx.asnumpy())

    b_mx = sockeye.layers.MultiHeadAttentionBase(hidden, heads, hidden)
    b_mx.initialize()
    b_pt = sockeye.layers_pt.PyTorchMultiHeadAttentionBase(hidden, heads, hidden)
    # use mxnet parameter initializations for pytorch block
    b_pt.ff_out.weight.data[:] = pt.as_tensor(b_mx.ff_out.weight.data().asnumpy())

    r_mx = b_mx._attend(q_mx, kv_mx, None, None).asnumpy()
    r_pt = b_pt._attend(q_pt, kv_pt, mask=None).detach().numpy()

    assert np.allclose(r_mx, r_pt, atol=1e-06)


@pytest.mark.parametrize('seq_len, batch_size, hidden, heads',
                         [(10, 1, 12, 4), (1, 1, 4, 1), (3, 15, 64, 2)])
def test_mx_pt_eq_multi_head_self_attention(seq_len, batch_size, hidden, heads):
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.layers

    inputs_mx = np.random.uniform(0, 1, (seq_len, batch_size, hidden))
    inputs_pt = pt.as_tensor(inputs_mx.asnumpy())

    b_mx = sockeye.layers.MultiHeadSelfAttention(hidden, heads, hidden, dropout=0.0)
    b_mx.initialize()
    b_pt = sockeye.layers_pt.PyTorchMultiHeadSelfAttention(hidden, heads, hidden, dropout=0.0)
    b_pt.weights_from_mxnet_block(b_mx)

    r_mx, states_mx = b_mx(inputs_mx, None, None, None)
    r_pt, states_pt = b_pt(inputs_pt, previous_states=None, inputs_length_mask=None, bias=None)

    r_mx = r_mx.asnumpy()
    states_mx = states_mx.asnumpy()
    r_pt = r_pt.detach().numpy()
    states_pt = states_pt.detach().numpy()

    assert np.allclose(r_mx, r_pt, atol=1e-06)
    assert np.allclose(states_mx, states_pt, atol=1e-06)


@pytest.mark.parametrize('qlen, kvlen, batch_size, hidden, heads',
                         [(10, 9, 1, 12, 4), (1, 1, 1, 4, 1), (3, 32, 15, 64, 2)])
def test_mx_pt_eq_multi_head_attention(qlen, kvlen, batch_size, hidden, heads):
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.layers

    queries_mx = np.random.uniform(0, 1, (qlen, batch_size, hidden))
    queries_pt = pt.as_tensor(queries_mx.asnumpy())
    memory_mx = np.random.uniform(0, 1, (kvlen, batch_size, hidden))
    memory_pt = pt.as_tensor(memory_mx.asnumpy())

    b_mx = sockeye.layers.MultiHeadAttention(hidden, heads, hidden, dropout=0.0)
    b_mx.initialize()
    r_mx = b_mx(queries_mx, memory_mx, None, None, None)

    b_pt = sockeye.layers_pt.PyTorchMultiHeadAttention(hidden, heads, hidden, dropout=0.0, depth_key_value=hidden)
    b_pt.weights_from_mxnet_block(b_mx)
    r_pt = b_pt(queries_pt, memory_pt, mask=None, projected_memory_kv=None)

    print(b_pt.ff_kv.weight[0])
    print(b_mx.ff_kv.weight.data()[0])

    r_mx = r_mx.asnumpy()
    r_pt = r_pt.detach().numpy()

    assert np.allclose(r_mx, r_pt, atol=1e-06)


@pytest.mark.parametrize('hidden, inference_only, seq_len, batch',
                         [(16, False, 10, 4),
                          (10, False, 2, 1),
                          (16, True, 1, 4), ])
def test_mx_pt_eq_ssru(hidden, inference_only, seq_len, batch):
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.layers

    b_mx = sockeye.layers.SSRU(hidden, inference_only)
    b_mx.initialize()
    b_pt = sockeye.layers_pt.PyTorchSSRU(hidden, inference_only)
    b_pt.weights_from_mxnet_block(b_mx)

    inputs_mx = np.random.uniform(0, 1, (seq_len, batch, hidden))
    previous_states_mx = np.zeros((1, batch, hidden))
    inputs_pt = pt.as_tensor(inputs_mx.asnumpy())
    previous_states_pt = pt.as_tensor(previous_states_mx.asnumpy())

    r1_mx, r2_mx = b_mx(inputs_mx, previous_states_mx)
    r1_pt, r2_pt = b_pt(inputs_pt, previous_states_pt)

    r1_mx = r1_mx.asnumpy()
    r2_mx = r2_mx.asnumpy()
    r1_pt = r1_pt.detach().numpy()
    r2_pt = r2_pt.detach().numpy()

    assert np.allclose(r1_mx, r1_pt)
    assert np.allclose(r2_mx, r2_pt)


def test_mx_pt_eq_length_ratio():
    pytest.importorskip("mxnet")
    from mxnet import np
    import sockeye.layers

    hidden_size = 32
    seq_len = 10
    batch_size = 8
    num_layers = 1  # more layers seems to be numerically unstable

    source_encoded_mx = np.random.uniform(0, 1, (batch_size, seq_len, hidden_size))
    source_encoded_pt = pt.as_tensor(source_encoded_mx.asnumpy())
    source_lengths_mx = np.random.randint(1, seq_len, (batch_size,), dtype='int32')
    source_lengths_pt = pt.as_tensor(source_lengths_mx.asnumpy())

    b_mx = sockeye.layers.LengthRatio(hidden_size=hidden_size, num_layers=num_layers)
    b_mx.initialize()
    r_mx = b_mx(source_encoded_mx, source_lengths_mx).asnumpy()

    b_pt = sockeye.layers_pt.PyTorchLengthRatio(hidden_size=hidden_size, num_layers=num_layers)
    b_pt.weights_from_mxnet_block(b_mx)
    r_pt = b_pt(source_encoded_pt, source_lengths_pt).detach().numpy()

    assert np.allclose(r_mx, r_pt)



# Copyright 2017--2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from abc import abstractmethod
from dataclasses import dataclass, field
from typing import List, Tuple, Optional, Union

import torch as pt

import sockeye.constants as C
from . import config
from . import layers_pt
from . import transformer_pt


def pytorch_get_transformer_encoder(config: transformer_pt.TransformerConfig, inference_only: bool = False):
    return PyTorchTransformerEncoder(config=config, inference_only=inference_only)


get_encoder = pytorch_get_transformer_encoder
EncoderConfig = Union[transformer_pt.TransformerConfig]


class PyTorchEncoder(pt.nn.Module):
    """
    Generic encoder interface.
    """

    @abstractmethod
    def get_num_hidden(self) -> int:
        """
        :return: The representation size of this encoder.
        """
        raise NotImplementedError()

    def get_encoded_seq_len(self, seq_len: int) -> int:
        """
        :return: The size of the encoded sequence.
        """
        return seq_len

    def get_max_seq_len(self) -> Optional[int]:
        """
        :return: The maximum length supported by the encoder if such a restriction exists.
        """
        return None


@dataclass
class FactorConfig(config.Config):
    vocab_size: int
    num_embed: int
    combine: str  # From C.FACTORS_COMBINE_CHOICES
    share_embedding: bool


@dataclass
class EmbeddingConfig(config.Config):
    vocab_size: int
    num_embed: int
    dropout: float
    num_factors: int = field(init=False)
    factor_configs: Optional[List[FactorConfig]] = None
    allow_sparse_grad: bool = False

    def __post_init__(self):
        self.num_factors = 1
        if self.factor_configs is not None:
            self.num_factors += len(self.factor_configs)


class PyTorchEmbedding(PyTorchEncoder):
    """
    Thin wrapper around PyTorch's Embedding op.

    :param config: Embedding config.
    :param embedding: pre-existing embedding Module.
    """

    def __init__(self, config: EmbeddingConfig, embedding: Optional[pt.nn.Embedding] = None) -> None:
        super().__init__()
        self.config = config

        if embedding is not None:
            self.embedding = embedding
        else:
            self.embedding = pt.nn.Embedding(self.config.vocab_size, self.config.num_embed,
                                             sparse=self.config.allow_sparse_grad)

        self.factor_embeds = pt.nn.ModuleList()
        if self.config.factor_configs is not None:
            for i, fc in enumerate(self.config.factor_configs, 1):
                if fc.share_embedding:
                    factor_embed = self.embedding
                else:
                    factor_embed = pt.nn.Embedding(fc.vocab_size, fc.num_embed,
                                                   sparse=self.config.allow_sparse_grad)
                self.factor_embeds.append(factor_embed)

        self.dropout = pt.nn.Dropout(p=self.config.dropout) if self.config.dropout > 0.0 else None

    def forward(self, data: pt.Tensor) -> Tuple[pt.Tensor, pt.Tensor]:  # pylint: disable=arguments-differ
        # We will catch the optional factor weights in kwargs
        average_factors_embeds = []  # type: List[pt.Tensor]
        concat_factors_embeds = []  # type: List[pt.Tensor]
        sum_factors_embeds = []  # type: List[pt.Tensor]

        if self.config.num_factors > 1 and self.config.factor_configs is not None:
            data, *factor_datas = (x.squeeze(2) for x in data.split(1, dim=2))
            for i, (factor_data, factor_embedding, factor_config) in enumerate(zip(factor_datas,
                                                                               self.factor_embeds,
                                                                               self.config.factor_configs)):
                factor_embedded = factor_embedding(factor_data)
                if factor_config.combine == C.FACTORS_COMBINE_CONCAT:
                    concat_factors_embeds.append(factor_embedded)
                elif factor_config.combine == C.FACTORS_COMBINE_SUM:
                    sum_factors_embeds.append(factor_embedded)
                elif factor_config.combine == C.FACTORS_COMBINE_AVERAGE:
                    average_factors_embeds.append(factor_embedded)
                else:
                    raise ValueError("Unknown combine value for factors: %s" % factor_config.combine)
        else:
            data = data.squeeze(2)

        embedded = self.embedding(data)

        if self.config.num_factors > 1 and self.config.factor_configs is not None:
            if average_factors_embeds:
                embedded = pt.mean(pt.stack([embedded] + average_factors_embeds, dim=0), dim=0)
            if sum_factors_embeds:
                embedded = pt.sum(pt.stack([embedded] + sum_factors_embeds, dim=0), dim=0)
            if concat_factors_embeds:
                embedded = pt.cat([embedded] + concat_factors_embeds, dim=2)

        if self.dropout is not None:
            embedded = self.dropout(embedded)

        return embedded

    def get_num_hidden(self) -> int:
        """
        Return the representation size of this encoder.
        """
        return self.config.num_embed

    def weights_from_mxnet_block(self, block_mx: 'Embedding'):  # type: ignore
        self.embedding.weight.data[:] = pt.as_tensor(block_mx.weight.data().asnumpy())
        if self.config.factor_configs is not None:
            for embedding, mx_weight, fc in zip(self.factor_embeds, block_mx.factor_weights, self.config.factor_configs):
                embedding.weight.data[:] = pt.as_tensor(mx_weight.data().asnumpy())


class PyTorchTransformerEncoder(PyTorchEncoder):
    """
    Non-recurrent encoder based on the transformer architecture in:

    Attention Is All You Need, Figure 1 (left)
    Vaswani et al. (https://arxiv.org/pdf/1706.03762.pdf).

    :param config: Configuration for transformer encoder.
    """

    def __init__(self, config: transformer_pt.TransformerConfig, inference_only: bool = False) -> None:
        pt.nn.Module.__init__(self)
        self.config = config

        self.dropout = pt.nn.Dropout(p=config.dropout_prepost) if config.dropout_prepost > 0.0 else None

        self.pos_embedding = layers_pt.PyTorchPositionalEmbeddings(weight_type=self.config.positional_embedding_type,
                                                                   num_embed=self.config.model_size,
                                                                   max_seq_len=self.config.max_seq_len_source,
                                                                   scale_up_input=True,
                                                                   scale_down_positions=False)

        self.layers = pt.nn.ModuleList(  # using ModuleList because we have additional inputs
            transformer_pt.PyTorchTransformerEncoderBlock(config, inference_only=inference_only)
            for _ in range(config.num_layers))

        self.final_process = transformer_pt.PyTorchTransformerProcessBlock(sequence=config.preprocess_sequence,
                                                                           dropout=config.dropout_prepost,
                                                                           num_hidden=self.config.model_size)

    def forward(self, data: pt.Tensor, valid_length: pt.Tensor) -> Tuple[pt.Tensor, pt.Tensor]:
        # positional embedding
        data = self.pos_embedding(data)

        if self.dropout is not None:
            data = self.dropout(data)

        # inverted length_mask for attention masking, (batch_size * heads, 1, max_len)
        att_mask = layers_pt.prepare_source_length_mask(valid_length, self.config.attention_heads, data.size()[1])

        data = data.transpose(1, 0)  # batch to time major
        for layer in self.layers:
            data = layer(data, att_mask=att_mask)

        data = self.final_process(data)
        data = data.transpose(1, 0)  # time to batch major
        return data, valid_length

    def get_num_hidden(self) -> int:
        """
        Return the representation size of this encoder.
        """
        return self.config.model_size

    def weights_from_mxnet_block(self, block_mx: 'TransformerEncoder'):  # type: ignore
        self.pos_embedding.weights_from_mxnet_block(block_mx.pos_embedding)
        for i, l in enumerate(self.layers):
            l.weights_from_mxnet_block(block_mx.layers[i])
        self.final_process.weights_from_mxnet_block(block_mx.final_process)

